"""The module :mod:`~pyrealm.pmodel.pmodel` provides the implementation of
the following  classes:

* :class:`~pyrealm.pmodel.pmodel.PModelABC`: An abstract base class providing some
  of the core functionality for initialising PModel subclasses.

* :class:`~pyrealm.pmodel.pmodel.PModel`: A subclass providing the standard
  implementation of the P Model.

* :class:`~pyrealm.pmodel.pmodel.SubdailyPModel`: A subclass providing the
  subdaily implementation of the P Model, which accounts for slow acclimation of core
  photosynthetic processes.


"""  # noqa D210, D415

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any
from warnings import warn

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import CoreConst, PModelConst
from pyrealm.core.utilities import summarize_attrs
from pyrealm.pmodel.acclimation import AcclimationModel
from pyrealm.pmodel.arrhenius import ARRHENIUS_METHOD_REGISTRY, ArrheniusFactorABC
from pyrealm.pmodel.jmax_limitation import (
    JMAX_LIMITATION_CLASS_REGISTRY,
    JmaxLimitationABC,
)
from pyrealm.pmodel.optimal_chi import OPTIMAL_CHI_CLASS_REGISTRY, OptimalChiABC
from pyrealm.pmodel.pmodel_environment import PModelEnvironment
from pyrealm.pmodel.quantum_yield import QUANTUM_YIELD_CLASS_REGISTRY, QuantumYieldABC


class PModelABC(ABC):
    r"""Abstract base class for the PModel and SubdailyPModel.

    The base class ``__init__`` implements the core arguments to the PModel subclasses:
    the forcing data to be used for the model and various methodological options for the
    calculation of the model parameters.

    Subclasses should define an ``__init__`` method that first calls
    ``super().__init__(...)`` to run the shared core functionality and then define any
    model specific attributes. The abstract base method ``_fit_model`` should then be
    defined and used to execute the model specific logic of the base class.

    Args:
        env: A :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment` instance
        method_kphio: The method to use for calculating the quantum yield
            efficiency of photosynthesis (:math:`\phi_0`, unitless). The method name
            must be included in the
            :data:`~pyrealm.pmodel.quantum_yield.QUANTUM_YIELD_CLASS_REGISTRY`.
        method_optchi: (Optional, default=`prentice14`) Selects the method to be
            used for calculating optimal :math:`chi`. The choice of method also sets the
            choice of  C3 or C4 photosynthetic pathway (see
            :class:`~pyrealm.pmodel.optimal_chi.OptimalChiABC`).
        method_jmaxlim: (Optional, default=`wang17`) Method to use for
            :math:`J_{max}` limitation.
        method_arrhenius: (Optional, default=`simple`) Method to set the form of
            Arrhenius scaling used for `vcmax` and `jmax`.
        reference_kphio: An optional alternative reference value for the quantum yield
            efficiency of photosynthesis (:math:`\phi_0`, -) to be passed to the kphio
            calculation method.
    """

    _data_attributes: tuple[tuple[str, str], ...]
    """Data attributes included in the summarize table."""

    def __init__(
        self,
        env: PModelEnvironment,
        method_kphio: str = "temperature",
        method_optchi: str = "prentice14",
        method_jmaxlim: str = "wang17",
        method_arrhenius: str = "simple",
        reference_kphio: float | NDArray | None = None,
        **kwargs: dict[str, Any],
    ):
        self.shape: tuple = env.shape
        """Records the common numpy array shape of array inputs."""

        # Store a reference to the photosynthetic environment and a direct
        # reference to the parameterisation
        self.env: PModelEnvironment = env
        """The PModelEnvironment used to fit the P Model."""
        self.pmodel_const: PModelConst = env.pmodel_const
        """The PModelConst instance used to create the model environment."""
        self.core_const: CoreConst = env.core_const
        """The CoreConst instance used to create the model environment."""

        # -----------------------------------------------------------------------
        # Optimal Chi method setup
        # -----------------------------------------------------------------------

        self.method_optchi: str
        """The method used to calculate optimal chi."""
        self._optchi_class: type[OptimalChiABC]
        """The OptimalChiABC subclass to be used in the model."""
        self.optchi: OptimalChiABC
        r"""The optimal chi (:math:`\chi`) calculations for the fitted P Model."""

        self._method_setter(
            method_value_attr="method_optchi",
            method_class_attr="_optchi_class",
            method_value=method_optchi,
            method_registry=OPTIMAL_CHI_CLASS_REGISTRY,
        )

        # Set the C4 status
        self.c4: bool = self._optchi_class.is_c4
        """Boolean flag showing if the optimal chi method approximates a C3 or C4
        pathway."""

        # -----------------------------------------------------------------------
        # Cuantum yield of photosynthesis (kphio) setup
        # -----------------------------------------------------------------------

        self.method_kphio: str
        """The method used to calculate kphio."""
        self._kphio_class: type[QuantumYieldABC]
        """The QuantumYieldABC subclass to be used in the model."""
        self.kphio: QuantumYieldABC
        r"""The quantum yield (:math:`\phi_0`) calculations for the fitted P Model."""

        self._method_setter(
            method_value_attr="method_kphio",
            method_class_attr="_kphio_class",
            method_value=method_kphio,
            method_registry=QUANTUM_YIELD_CLASS_REGISTRY,
        )

        self.reference_kphio = reference_kphio
        """The value of the the reference kphio to be used in the model."""

        # -----------------------------------------------------------------------
        # Arrhenius scaling setup
        # -----------------------------------------------------------------------

        self.method_arrhenius = method_arrhenius
        """The method used to calculate Arrhenius factors."""

        self._arrhenius_class: type[ArrheniusFactorABC]
        """The ArrheniusFactorABC subclass to be used in the model."""

        self._method_setter(
            method_value_attr="method_arrhenius",
            method_class_attr="_arrhenius_class",
            method_value=method_arrhenius,
            method_registry=ARRHENIUS_METHOD_REGISTRY,
        )

        if self.method_arrhenius != "simple":
            warn(
                "We currently strongly recommend the use of the default `simple` "
                "method calculating Arrhenius factors."
            )

        # -----------------------------------------------------------------------
        # Jmax limitation term setup
        # -----------------------------------------------------------------------

        self.method_jmaxlim = method_jmaxlim
        """Records the method used to calculate Jmax limitation."""
        self._jmaxlim_class: type[JmaxLimitationABC]
        """The ArrheniusFactorABC subclass to be used in the model."""
        self.jmaxlim: JmaxLimitationABC
        """The Jmax limitation terms calculated for the model."""

        self._method_setter(
            method_value_attr="method_jmaxlim",
            method_class_attr="_jmaxlim_class",
            method_value=method_jmaxlim,
            method_registry=JMAX_LIMITATION_CLASS_REGISTRY,
        )

        # -----------------------------------------------------------------------
        # Define the other model attributes
        # -----------------------------------------------------------------------

        self.iwue: NDArray[np.float64]
        """Intrinsic water use efficiency (iWUE, µmol mol-1), calculated as:

        .. math::
        
            ( 5/8 * (c_a - c_i)) / P,
        
        where :math:`c_a` and :math:`c_i` are measured in Pa and :math:`P` is
        atmospheric pressure in megapascals.
        """

        self.lue: NDArray[np.float64]
        r"""Light use efficiency (LUE, g C mol-1), calculated as:

        .. math::

            \text{LUE} = \phi_0 \cdot m_j \cdot f_v \cdot M_C

        where :math:`f_v` is a limitation factor defined in
        :class:`~pyrealm.pmodel.jmax_limitation.JmaxLimitationABC` and :math:`M_C` is
        the molar mass of carbon.
        """

        self.vcmax: NDArray[np.float64]
        r"""Maximum rate of carboxylation at the growth temperature (µmol m-2 s-1),
        calculated as:

        .. math::

            V_{cmax} = \phi_{0} I_{abs} \frac{m}{m_c} f_{v} 

        where  :math:`f_v` is a limitation term calculated via the method selected in
        `method_jmaxlim`."""

        self.vcmax25: NDArray[np.float64]
        """Maximum rate of carboxylation at standard temperature (µmol m-2 s-1),
        estimated from :math:`V_{cmax}` using the selected method for Arrhenius scaling.
        """

        self.jmax: NDArray[np.float64]
        r"""Maximum rate of electron transport at the growth temperature (µmol m-2 s-1),
        calculated as:
        
        .. math::

            J_{max} = 4 \phi_{0} I_{abs} f_{j}

        where  :math:`f_j` is a limitation term calculated via the method selected in
        `method_jmaxlim`."""

        self.jmax25: NDArray[np.float64]
        """Maximum rate of electron transport at standard temperature (µmol m-2 s-1),
        estimated from :math:`J_{max}` using the selected method for Arrhenius scaling.
        """

        self.gpp: NDArray[np.float64]
        r"""Gross primary productivity (µg C m-2 s-1) calculated as:
        
        .. math::

            \text{GPP} = \text{LUE} \cdot I_{abs}
            
        where :math:`I_{abs}` is the absorbed photosynthetic radiation.
        """

        self.gs: NDArray[np.float64]
        r"""Stomatal conductance (µmol m-2 s-1), calculated as:

        .. math::

            g_s = \frac{LUE}{M_C}\frac{1}{c_a - c_i}

        When C4 photosynthesis is being used, the true partial pressure of CO2 in the
        substomatal cavities (:math:`c_i`) is used following the calculation of
        :math:`\chi` using
        :attr:`~pyrealm.constants.pmodel_const.PModelConst.beta_cost_ratio_c4`. Note
        that :math:`g_s \to \infty` as VPD :math:`\to 0` and hence :math:`(c_a - c_i)
        \to 0` and the reported values will be set to ``np.nan`` under these
        conditions."""

        self.A_c: NDArray[np.float64]
        """Maximum assimilation rate limited by carboxylation."""
        self.A_j: NDArray[np.float64]
        """Maximum assimilation rate limited by electron transport."""
        self.J: NDArray[np.float64]
        """Electron transfer rate."""

    @abstractmethod
    def _fit_model(self, *args: Any, **kwargs: Any) -> None:
        pass

    def __repr__(self) -> str:
        """Generates a string representation of PModel instance."""

        return (
            f"{self.__class__.__name__}("
            f"shape={self.shape}, "
            f"method_optchi={self.method_optchi}, "
            f"method_arrhenius={self.method_arrhenius}, "
            f"method_jmaxlim={self.method_jmaxlim}, "
            f"method_kphio={self.method_kphio})"
        )

    def summarize(self, dp: int = 2) -> None:
        """Prints a summary of data attributes.

        Args:
            dp: The number of decimal places used in rounding summary stats.
        """

        summarize_attrs(self, self._data_attributes, dp=dp)

    def _method_setter(
        self,
        method_value_attr: str,
        method_class_attr: str,
        method_value: str,
        method_registry: dict,
    ) -> None:
        """Validate and set the method class attributes for a PModel method argument.

        This method checks that the provided method value is an option in the registry
        and then stores both the value and the appropriate method implementation class
        from the registry.

        Args:
            method_value_attr: The name of the attribute used to store the selected
                method name (e.g. ``method_optchi``)
            method_class_attr: The name of the attribute used to store the resulting
                selected class that implements the method
            method_value: The value passed to the method selection argument
            method_registry: A method registry, providing method class implementations
                keyed by method name values.
        """
        if method_value not in method_registry:
            raise ValueError(f"Unknown option for {method_value_attr}: {method_value}")

        setattr(self, method_value_attr, method_value)
        setattr(self, method_class_attr, method_registry[method_value])


class PModel(PModelABC):
    r"""Fit a standard P Model.

    This class fits the P Model to a given set of environmental and photosynthetic
    parameters. An extended description with typical use cases is given in
    :any:`pmodel_overview` but the basic flow of the model is:

    1. Estimate :math:`\ce{CO2}` limitation factors and optimal internal to ambient
       :math:`\ce{CO2}` partial pressure ratios (:math:`\chi`), using one of the
       methods based on :class:`~pyrealm.pmodel.optimal_chi.OptimalChiABC`.
    2. Estimate limitation factors to :math:`V_{cmax}` and :math:`J_{max}` using
       one of the methods implemented using
       :class:`~pyrealm.pmodel.jmax_limitation.JmaxLimitationABC`.

    Soil moisture effects:
        The `lavergne20_c3`, `lavergne20_c4`, ``prentice14_rootzonestress``,
        ``c4_rootzonestress`` and ``c4_no_gamma_rootzonestress`` options to
        ``method_optchi`` implement different approaches to soil moisture effects on
        photosynthesis. See also the alternative GPP penalty factors that can be applied
        after fitting the P Model
        (:func:`pyrealm.pmodel.functions.calc_soilmstress_stocker` and
        :func:`pyrealm.pmodel.functions.calc_soilmstress_mengoli`).

    Args:
        env: An instance of
           :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment`
        method_kphio: The method to use for calculating the quantum yield
            efficiency of photosynthesis (:math:`\phi_0`, unitless). The method name
            must be included in the
            :data:`~pyrealm.pmodel.quantum_yield.QUANTUM_YIELD_CLASS_REGISTRY`.
        method_optchi: (Optional, default=`prentice14`) Selects the method to be
            used for calculating optimal :math:`chi`. The choice of method also sets the
            choice of  C3 or C4 photosynthetic pathway (see
            :class:`~pyrealm.pmodel.optimal_chi.OptimalChiABC`).
        method_jmaxlim: (Optional, default=`wang17`) Method to use for
            :math:`J_{max}` limitation.
        method_arrhenius: (Optional, default=`simple`) Method to set the form of
            Arrhenius scaling used for `vcmax` and `jmax`.
        reference_kphio: An optional alternative reference value for the quantum yield
            efficiency of photosynthesis (:math:`\phi_0`, -) to be passed to the kphio
            calculation method.
    """

    _data_attributes = (
        ("lue", "g C mol-1"),
        ("iwue", "µmol mol-1"),
        ("gpp", "µg C m-2 s-1"),
        ("vcmax", "µmol m-2 s-1"),
        ("vcmax25", "µmol m-2 s-1"),
        ("gs", "µmol m-2 s-1"),
        ("jmax", "µmol m-2 s-1"),
        ("jmax25", "µmol m-2 s-1"),
    )

    def __init__(
        self,
        env: PModelEnvironment,
        method_optchi: str = "prentice14",
        method_jmaxlim: str = "wang17",
        method_kphio: str = "temperature",
        method_arrhenius: str = "simple",
        reference_kphio: float | NDArray | None = None,
    ) -> None:
        # Initialise the superclass
        super().__init__(
            env=env,
            method_optchi=method_optchi,
            method_jmaxlim=method_jmaxlim,
            method_kphio=method_kphio,
            method_arrhenius=method_arrhenius,
            reference_kphio=reference_kphio,
        )
        # Fit the model
        self._fit_model()

    def _fit_model(self) -> None:
        """Fit the model.

        PPFD _must_ be provided in units of micromols per metre square per second (µmol
        m-2 s-1). This is required to ensure that values of :math:`J_{max}` and
        :math:`V_{cmax}` are also in µmol m-2 s-1.
        """

        warn(
            """
    Pyrealm 2.0.0 uses a new default for the quantum yield of photosynthesis (phi0=1/8).
    You may need to change settings to duplicate results from pyrealm 1.0.0.
            """,
            category=UserWarning,
        )

        # Calculate optimal chi
        self.optchi: OptimalChiABC = self._optchi_class(
            env=self.env,
            pmodel_const=self.env.pmodel_const,
        )
        self.c4 = self.optchi.is_c4

        # Calculate the quantum yield of photosynthesis (kphio)
        self.kphio = self._kphio_class(
            env=self.env,
            use_c4=self.c4,
            reference_kphio=self.reference_kphio,
        )

        # Calculation of Jmax limitation terms
        self.jmaxlim = self._jmaxlim_class(
            optchi=self.optchi,
            pmodel_const=self.env.pmodel_const,
        )

        # Get an Arrhenius instance for use in scaling rates
        arrhenius_factors = self._arrhenius_class(env=self.env)

        # Intrinsic water use efficiency (in µmol mol-1)
        self.iwue = (5 / 8 * (self.env.ca - self.optchi.ci)) / (1e-6 * self.env.patm)

        # Light use efficiency g Carbon per mol-1 of photons.
        self.lue = (
            self.kphio.kphio
            * self.optchi.mj
            * self.jmaxlim.f_v
            * self.core_const.k_c_molmass
        )

        # Calculate absorbed irradiance
        iabs = self.env.fapar * self.env.ppfd

        # GPP
        self.gpp = self.lue * iabs

        # Calculate V_cmax and J_max
        self.vcmax = self.kphio.kphio * iabs * self.optchi.mjoc * self.jmaxlim.f_v
        self.jmax = 4 * self.kphio.kphio * iabs * self.jmaxlim.f_j

        # - Calculate and apply the scaling factors.
        ftemp25_inst_vcmax = arrhenius_factors.calculate_arrhenius_factor(
            coefficients=self.pmodel_const.arrhenius_vcmax
        )
        self.vcmax25 = self.vcmax / ftemp25_inst_vcmax
        self.jmax25 = self.jmax / arrhenius_factors.calculate_arrhenius_factor(
            coefficients=self.pmodel_const.arrhenius_jmax
        )

        # AJ and AC
        self.A_j = self.kphio.kphio * iabs * self.optchi.mj * self.jmaxlim.f_v
        self.A_c = self.vcmax * self.optchi.mc

        if not np.allclose(self.A_j, self.A_c, equal_nan=True):
            raise RuntimeError(
                "Violation of coordination hypothesis: A_c is not equal to A_j"
            )

        # Stomatal conductance - do not estimate when VPD = 0 or when floating point
        # errors give rise to (ca - ci) < 0 and deliberately ignore the numpy divide by
        # zero warnings in those cases.
        ca_ci_diff = self.env.ca - self.optchi.ci
        with np.errstate(divide="ignore", invalid="ignore"):
            self.gs = np.where(
                np.logical_and(self.env.vpd > 0, ca_ci_diff > 0),
                self.A_c / ca_ci_diff,
                np.nan,
            )

    def to_subdaily(
        self,
        acclim_model: AcclimationModel,
        previous_realised: tuple[NDArray, NDArray, NDArray] | None = None,
    ) -> SubdailyPModel:
        r"""Convert a standard PModel to a subdaily P Model.

        This method converts a :class:`~pyrealm.pmodel.pmodel.PModel` instance to
        a to a :class:`~pyrealm.pmodel.pmodel.SubdailyPModel` instance with the
        same settings.

        Args:
            acclim_model: An AcclimationModel instance for the subdaily model.
            previous_realised: An optional set of arrays giving previous realised values
                for `xi`, `vcmax25` and `jmax25`.
        """

        return SubdailyPModel(
            env=self.env,
            method_optchi=self.method_optchi,
            method_arrhenius=self.method_arrhenius,
            method_jmaxlim=self.method_jmaxlim,
            method_kphio=self.method_kphio,
            reference_kphio=self.kphio.reference_kphio,
            acclim_model=acclim_model,
        )


class SubdailyPModel(PModelABC):
    r"""Fit a P Model incorporating acclimation in photosynthetic responses.

    The :class:`~pyrealm.pmodel.pmodel.PModel` implementation of the P Model
    assumes that plants instantaneously adopt optimal behaviour, which is reasonable
    where the data represents average conditions over longer timescales and the plants
    can be assumed to have acclimated to optimal behaviour. Over shorter timescales,
    this assumption is unwarranted and photosynthetic slow responses need to be
    included. This class implements the weighted-average approach of
    {cite:t}`mengoli:2022a`, but is extended to include the slow response of :math:`\xi`
    in addition to :math:`V_{cmax25}` and :math:`J_{max25}`.

    The workflow of the model:

    * The first dimension of the data arrays used to create the
      :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment` instance must
      represent the time axis of the observations. The ``acclim_model`` argument is used
      to provide a :class:`~pyrealm.pmodel.acclimation.AcclimationModel` instance that
      sets the dates and time of those observations. One of the ``set_`` methods to that
      class must also be used to define a daily acclimation window that will be used to
      estimate the optimal daily behaviour of the plant.
    * The
      :meth:`AcclimationModel.get_daily_means<pyrealm.pmodel.acclimation.AcclimationModel.get_daily_means>`
      method is then used to extract daily average values for forcing variables from
      within the acclimation window, setting the conditions that the plant will optimise
      to.
    * A standard P Model is then run on those daily forcing values to generate predicted
      states for photosynthetic parameters that give rise to optimal productivity in
      that window.
    * The
      :meth:`AcclimationModel.apply_acclimation<pyrealm.pmodel.acclimation.AcclimationModel.apply_acclimation>`
      method is then used to calculate acclimating values for
      :math:`\xi`, :math:`V_{cmax25}` and :math:`J_{max25}`. These values are the actual
      realised values that will be used in calculating the P Model and which reflect the
      slow responses of those parameters to changing conditions. The speed of
      acclimation is controlled by the
      :attr:`AcclimationModel.alpha<pyrealm.pmodel.acclimation.AcclimationModel.alpha>`
      attribute, which sets the weight :math:`\alpha \in [0,1]` for an exponential
      moving average (see :meth:`pyrealm.core.utilities.exponential_moving_average`)
      Higher values of `alpha` give more rapid acclimation: :math:`\alpha=1`
      results in immediate acclimation and :math:`\alpha=0` results in no acclimation at
      all, with values pinned to the initial estimates.
    * By default, the initial values of the acclimated variables are taken to be the
      same as the initial optimal values. The `previous_realised` argument can be used
      to provide alternative initial realised values. This allows subdaily P models to
      be `restarted` using estimates of acclimated values.
    * The realised values are then filled back onto the original subdaily timescale,
      using the
      :meth:`AcclimationModel.fill_daily_to_subdaily<pyrealm.pmodel.acclimation.AcclimationModel.fill_daily_to_subdaily>`
      method. The subdaily values of :math:`V_{cmax}` and :math:`J_{max}` are calculated
      by using Arrhenius scaling to convert the acclimated values of :math:`V_{cmax25}`
      and :math:`J_{max25}` to the actual subdaily
      temperature observations. The value of :math:`c_i` is calculated using the
      acclimating values of :math:`\xi` but the actual subdaily values of temperature
      and vapour pressure deficit.
    * Predictions of GPP are then made as in the standard P Model.

    As with the :class:`~pyrealm.pmodel.pmodel.PModel`, the values of the `kphio`
    argument _can_ be provided as an array of values, potentially varying through time
    and space. The behaviour of the daily model that drives acclimation here is to take
    the daily mean `kphio` value for each time series within the acclimation window, as
    for the other variables. This is an experimental solution!

    Missing values:

        Missing data can arise in a number of ways: actual gaps in the forcing data, the
        observations starting part way through a day and missing some or all of the
        acclimation window for the day, or undefined values in P Model predictions. Some
        options include:

        * The ``allow_partial_data`` argument is passed on to the
          :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.get_daily_means` method to
          allow daily optimum conditions to be calculated when the data in the
          acclimation window is incomplete. This does not fix problems when no data is
          present in the window or when the P Model predictions for a day are undefined.

        * The ``allow_holdover`` argument is passed on to the
          :meth:`~pyrealm.pmodel.acclimation.AcclimationModel.apply_acclimation` method
          to set whether missing values in the optimal predictions can be filled by
          holding over previous valid values.

    Args:
        env: An instance of
           :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment`
        acclim_model: An instance of
            :class:`~pyrealm.pmodel.acclimation.AcclimationModel`
        method_kphio: The method to use for calculating the quantum yield
            efficiency of photosynthesis (:math:`\phi_0`, unitless). The method name
            must be included in the
            :data:`~pyrealm.pmodel.quantum_yield.QUANTUM_YIELD_CLASS_REGISTRY`.
        method_optchi: (Optional, default=`prentice14`) Selects the method to be
            used for calculating optimal :math:`chi`. The choice of method also sets the
            choice of  C3 or C4 photosynthetic pathway (see
            :class:`~pyrealm.pmodel.optimal_chi.OptimalChiABC`).
        method_jmaxlim: (Optional, default=`wang17`) Method to use for
            :math:`J_{max}` limitation.
        method_arrhenius: (Optional, default=`simple`) Method to set the form of
            Arrhenius scaling used for `vcmax` and `jmax`.
        reference_kphio: An optional alternative reference value for the quantum yield
            efficiency of photosynthesis (:math:`\phi_0`, -) to be passed to the kphio
            calculation method.
        previous_realised: A tuple of previous realised values of three NumPy arrays
            (xi_real, vcmax25_real, jmax25_real).
    """

    _data_attributes = (
        ("iwue", "µmol mol-1"),
        ("gpp", "µg C m-2 s-1"),
        ("vcmax", "µmol m-2 s-1"),
        ("vcmax25", "µmol m-2 s-1"),
        ("gs", "µmol m-2 s-1"),
        ("jmax", "µmol m-2 s-1"),
        ("jmax25", "µmol m-2 s-1"),
    )

    def __init__(
        self,
        env: PModelEnvironment,
        acclim_model: AcclimationModel,
        method_optchi: str = "prentice14",
        method_jmaxlim: str = "wang17",
        method_kphio: str = "temperature",
        method_arrhenius: str = "simple",
        reference_kphio: float | NDArray | None = None,
        previous_realised: dict[str, NDArray] | None = None,
    ) -> None:
        # Initialise the superclass
        super().__init__(
            env=env,
            method_optchi=method_optchi,
            method_jmaxlim=method_jmaxlim,
            method_kphio=method_kphio,
            method_arrhenius=method_arrhenius,
            reference_kphio=reference_kphio,
        )

        # Subclass specific attributes
        self.acclim_model: AcclimationModel
        """The acclimation model used in the subdaily P Model."""
        self.previous_realised: Mapping[str, NDArray | None]
        """A dictionary of arrays of previous realised values for the acclimating
        variables 'xi', 'jmax25' and 'vcmax25'. If none were provided, the dictionary
        values are None."""

        # Other attributes
        self.pmodel_acclim: PModel
        r"""P Model predictions for the daily acclimation conditions.

        A :class:`~pyrealm.pmodel.pmodel.PModel` instance providing the
        predictions of the P Model for the daily acclimation conditions set for the
        SubdailyPModel. The model is used to obtain predictions of the instantaneous
        optimal estimates of :math:`V_{cmax}`, :math:`J_{max}` and :math:`\xi` during
        the acclimation window. These are then used to estimate realised values of those
        parameters given slow responses to acclimation.
        """

        # TODO - maybe encapsulate these in a dataclass?
        self.vcmax25_daily_optimal: NDArray[np.float64]
        r"""Daily optimal values in acclimation window for :math:`V_{cmax}`, scaled to
         standard temperature (:math:`V_{cmax25}`)."""
        self.vcmax25_daily_realised: NDArray[np.float64]
        r"""Realised daily responses in :math:`V_{cmax25}`"""
        self.jmax25_daily_optimal: NDArray[np.float64]
        r"""Daily optimal values in acclimation window for :math:`J_{max}`, scaled to
         standard temperature (:math:`J_{max25}`)."""
        self.jmax25_daily_realised: NDArray[np.float64]
        r"""Realised daily responses in :math:`J_{max25}`"""
        self.xi_daily_optimal: NDArray[np.float64]
        r"""Daily optimal values in acclimation window for :math:`\xi`"""
        self.xi_daily_realised: NDArray[np.float64]
        r"""Realised daily responses in :math:`\xi`"""

        # Fit the model
        self._fit_model(acclim_model=acclim_model, previous_realised=previous_realised)

    def _fit_model(
        self,
        acclim_model: AcclimationModel,
        previous_realised: dict[str, NDArray] | None,
    ) -> None:
        """Calculation logic of the subdaily P Model."""

        # Validate subdaily model specific arguments
        # * Check that the number of datetimes in the AcclimationModel is the same as
        #   the length of the first axis of the photosynthetic environment and that the
        #   one of the set_ methods has been run on the AcclimationModel.

        # Store the acclimation model
        self.acclim_model = acclim_model

        if self.acclim_model.datetimes.shape[0] != self.env.tc.shape[0]:
            raise ValueError(
                "The PModelEnvironment data and AcclimationModel datetimes "
                "are of different lengths."
            )

        if not hasattr(self.acclim_model, "include"):
            raise ValueError(
                "The daily sampling window has not been set in the AcclimationModel"
            )
        # * Validate the previous realised values and standardise the internal
        #   representation as a dictionary.
        if previous_realised is None:
            self.previous_realised = {"xi": None, "jmax25": None, "vcmax25": None}
        else:
            # Is the fill method set to previous
            if self.acclim_model.fill_method != "previous":
                raise NotImplementedError(
                    "Using previous_realised is only implemented for "
                    "fill_method = 'previous'"
                )

            # Check it is a dictionary of numpy arrays for the three required variables
            if not (
                isinstance(previous_realised, dict)
                and (set(["xi", "jmax25", "vcmax25"]) == previous_realised.keys())
                and all(
                    [isinstance(val, np.ndarray) for val in previous_realised.values()]
                )
            ):
                raise ValueError(
                    "previous_realised must be a dictionary of arrays, with entries "
                    "for 'xi', 'jmax25' and 'vcmax25'."
                )

            # All variables should share the shape of a slice along the first axis of
            # the environmental forcings. Need to tell mypy to shut up - it does not
            # know that the values in previous_realised are confirmed to be arrays by
            # the code above

            try:
                for values in previous_realised.values():
                    _ = np.broadcast_shapes(self.env.tc.shape, values.shape)
            except ValueError:
                raise ValueError(
                    "`previous_realised` arrays have wrong shape in SubdailyPModel"
                )

            self.previous_realised = previous_realised

        # 1) Generate a PModelEnvironment containing the average conditions within the
        #    daily acclimation window. This daily average environment also needs to also
        #    pass through any optional variables required by the optimal chi and kphio
        #    method set for the model, which can be accessed via the class requires
        #    attribute.

        # Get the list of variables for which to calculate daily acclimation conditions.
        daily_environment_vars = [
            "tc",
            "co2",
            "patm",
            "vpd",
            "fapar",
            "ppfd",
            *self._optchi_class.requires,
            *self._kphio_class.requires,
        ]

        # Construct a dictionary of daily acclimation variables, handling optional
        # choices which can be None.
        daily_environment: dict[str, NDArray] = {}
        for env_var_name in daily_environment_vars:
            env_var = getattr(self.env, env_var_name)
            if env_var is not None:
                daily_environment[env_var_name] = self.acclim_model.get_daily_means(
                    values=env_var,
                )

        # Calculate the acclimation environment passing on the constants definitions.
        pmodel_env_acclim: PModelEnvironment = PModelEnvironment(
            **daily_environment,
            bounds_checker=self.env._bounds_checker,
            pmodel_const=self.env.pmodel_const,
            core_const=self.env.core_const,
        )

        # Handle the kphio settings. First, calculate kphio at the subdaily scale.
        self.kphio: QuantumYieldABC = self._kphio_class(
            env=self.env,
            use_c4=self.c4,
            reference_kphio=self.reference_kphio,
        )

        # If the kphio method takes a single reference value then we can simply
        # recalculate the kphio using the same method for the daily acclimation
        # conditions but if the reference value is an array then the correct behaviour
        # is not obvious: currently, use the mean calculated kphio within the window to
        # calculate the daily acclimation value behaviour and set the kphio method to be
        # fixed to avoid altering the inputs.
        if self.kphio.reference_kphio.size > 1:
            daily_reference_kphio = self.acclim_model.get_daily_means(self.kphio.kphio)
            daily_method_kphio = "fixed"
        else:
            daily_reference_kphio = self.kphio.reference_kphio
            daily_method_kphio = self.method_kphio

        # 2) Fit a PModel to those environmental conditions, using the supplied settings
        #    for the original model.
        self.pmodel_acclim = PModel(
            env=pmodel_env_acclim,
            method_kphio=daily_method_kphio,
            method_optchi=self.method_optchi,
            method_jmaxlim=self.method_jmaxlim,
            method_arrhenius=self.method_arrhenius,
            reference_kphio=daily_reference_kphio,
        )
        self.pmodel_acclim._fit_model()

        # 4) Calculate the daily optimal values. Xi is simply the value from the optimal
        #   chi calculation but jmax and vcmax are scaled to values at 25°C using an
        #   instance of the requested Arrhenius scaling method .

        self.xi_daily_optimal = self.pmodel_acclim.optchi.xi

        arrhenius_daily = self._arrhenius_class(env=self.pmodel_acclim.env)

        self.vcmax25_daily_optimal = (
            self.pmodel_acclim.vcmax
            / arrhenius_daily.calculate_arrhenius_factor(
                coefficients=self.env.pmodel_const.arrhenius_vcmax
            )
        )
        self.jmax25_daily_optimal = (
            self.pmodel_acclim.jmax
            / arrhenius_daily.calculate_arrhenius_factor(
                coefficients=self.env.pmodel_const.arrhenius_jmax
            )
        )

        # 5) Calculate the realised daily values from the instantaneous optimal values

        self.xi_daily_realised = self.acclim_model.apply_acclimation(
            values=self.xi_daily_optimal,
            initial_values=self.previous_realised["xi"],
        )

        self.vcmax25_daily_realised = self.acclim_model.apply_acclimation(
            values=self.vcmax25_daily_optimal,
            initial_values=self.previous_realised["vcmax25"],
        )

        self.jmax25_daily_realised = self.acclim_model.apply_acclimation(
            values=self.jmax25_daily_optimal,
            initial_values=self.previous_realised["jmax25"],
        )

        # 6) Fill the realised xi, jmax25 and vcmax25 from daily values back to the
        # subdaily timescale.
        self.xi = self.acclim_model.fill_daily_to_subdaily(
            values=self.xi_daily_realised,
            previous_values=self.previous_realised["xi"],
        )
        self.vcmax25 = self.acclim_model.fill_daily_to_subdaily(
            self.vcmax25_daily_realised,
            previous_values=self.previous_realised["vcmax25"],
        )
        self.jmax25 = self.acclim_model.fill_daily_to_subdaily(
            self.jmax25_daily_realised,
            previous_values=self.previous_realised["jmax25"],
        )

        # 7) Adjust subdaily jmax25 and vcmax25 back to jmax and vcmax given the
        #    actual subdaily temperatures.
        arrhenius_subdaily = self._arrhenius_class(env=self.env)

        self.vcmax: NDArray[np.float64] = (
            self.vcmax25
            * arrhenius_subdaily.calculate_arrhenius_factor(
                coefficients=self.env.pmodel_const.arrhenius_vcmax
            )
        )

        self.jmax: NDArray[np.float64] = (
            self.jmax25
            * arrhenius_subdaily.calculate_arrhenius_factor(
                coefficients=self.env.pmodel_const.arrhenius_jmax
            )
        )

        # 8) Recalculate chi, but using the realised values of the xi parameter at the
        #    subdaily scale, not the instantaneous values of xi.
        self.optchi = self._optchi_class(
            env=self.env, pmodel_const=self.env.pmodel_const
        )
        self.optchi.estimate_chi(xi_values=self.xi)

        # Calculate Ac, J and Aj at subdaily scale to calculate assimilation
        self.A_c = self.vcmax * self.optchi.mc

        iabs = self.env.fapar * self.env.ppfd

        self.J = (4 * self.kphio.kphio * iabs) / np.sqrt(
            1 + ((4 * self.kphio.kphio * iabs) / self.jmax) ** 2
        )

        self.A_j = (self.J / 4) * self.optchi.mj

        # Calculate GPP and convert from mol to gC
        assimilation = np.minimum(self.A_j, self.A_c)
        self.gpp = assimilation * self.env.core_const.k_c_molmass

        # Stomatal conductance - do not estimate when VPD = 0 or when floating point
        # errors give rise to (ca - ci) < 0 and deliberately ignore the numpy divide by
        # zero warnings in those cases.
        ca_ci_diff = self.env.ca - self.optchi.ci
        with np.errstate(divide="ignore", invalid="ignore"):
            self.gs = np.where(
                np.logical_and(self.env.vpd > 0, ca_ci_diff > 0),
                self.A_c / ca_ci_diff,
                np.nan,
            )

        # Intrinsic water use efficiency (in µmol mol-1)
        self.iwue = (5 / 8 * (self.env.ca - self.optchi.ci)) / (1e-6 * self.env.patm)

    @property
    def lue(self) -> None:  # type: ignore[override]
        """The Subdaily P Model does not predict light use efficiency."""
        raise AttributeError(
            "The Subdaily P Model does not predict light use efficiency."
        )



"""A module to provide annotation tools for experimental features.

These tools should be applied to all tools that are not guaranteed to have a stable API
within major releases. The intention is that this functionality should:

* raise a warning when experimental features are used, and
* mark experimental features clearly in the documentation.

In an ideal world, we would use a decorator to do both of these but there is not
currently a simple implementation that preserves class signatures (see
https://github.com/ImperialCollegeLondon/pyrealm/issues/429). So for the moment, this
module provides the warning and the documentation is achieved using a custom sphinx
extension that looks for an `__experimental__ = True` attribute on objects. It is easy
to set a class variable for experimental classes but for experimental functions and
methods, we need to annotate the function or method separately (and tell ``mypy`` to
shut up).

So using  `__experimental__` looks like this:

.. code:: python

    def experimental_func(a: int = 1) -> 1:
        '''A docstring.'''

        warn_experimental(experimental_func)

        return a + 1

    experimental_func.__experimental__ = True  # type: ignore[attr-defined]


    class ExperimentalMethod:

        __experimental__ = True

        def experimental_method():

            pass

        experimental_method.__experimental__ = True  # type: ignore[attr-defined]

The ``warn_experimental`` function should be called from within the method or function
or from within the ``__init__`` for class instances.
"""

from warnings import warn


class ExperimentalFeatureWarning(Warning):
    """Warn about experimental features.

    This is just a simple wrapper on the base Warning to issue clearer warnings about
    experimental features in the code.

    Arg:
        qualname: The object name to add to the warning.
    """

    def __init__(self, qualname: str) -> None:
        self.message = (
            f"Be aware that {qualname} is an experimental feature of pyrealm and "
            f"the implementation and API may change within major versions."
        )

    def __str__(self) -> str:
        return repr(self.message)


def warn_experimental(qualname: str) -> None:
    """A simple wrapper function to generate an experimental warning."""

    warn(qualname, ExperimentalFeatureWarning)


def create_experimental_docstring(qualname: str) -> str:
    """Generate an RST admonition to indicate experimental usage.

    This cannot actually be used in practice because you cannot create a docstring
    dynamically - it has to be a single fixed string literal.

    Args:
        qualname: The object name to add to the admonition.
    """

    return f"""
.. admonition:: Experimental
    :class: Important
    
    The {qualname} method or class is an experimental feature and may
    change between major releases.
    """



r"""The module :mod:`~pyrealm.pmodel.quantum_yield` provides the abstract base class
:class:`~pyrealm.pmodel.quantum_yield.QuantumYieldABC`, which is used to support
different implementations of the calculation of the intrinsic quantum yield efficiency
of photosynthesis (:math:`\phi_0`, unitless). The module then provides subclasses of the
ABC implementing different approaches.

Note that :math:`\phi_0` is sometimes used to refer to the quantum yield of electron
transfer, which is exactly four times larger, so check definitions here.
"""  # noqa D210, D415

from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
from numpy.typing import NDArray

from pyrealm.core.experimental import warn_experimental
from pyrealm.core.utilities import (
    check_input_shapes,
    evaluate_horner_polynomial,
    summarize_attrs,
)
from pyrealm.pmodel.functions import calculate_kattge_knorr_arrhenius_factor
from pyrealm.pmodel.pmodel_environment import PModelEnvironment

QUANTUM_YIELD_CLASS_REGISTRY: dict[str, type[QuantumYieldABC]] = {}
r"""A registry for intrinsic quantum yield of photosynthesis calculation classes.

Different implementations of the calculation of the intrinsic quantum yield of
photosynthesis (:math:`\phi_{0}`) must all be subclasses of
:class:`~pyrealm.pmodel.quantum_yield.QuantumYieldABC` abstract base class. This
dictionary is used as a registry for defined subclasses and a method name is used to
retrieve a particular implementation from this registry. For example:

.. code:: python

    temperature_phio = QUANTUM_YIELD_CLASS_REGISTRY['temperature']
"""


class QuantumYieldABC(ABC):
    r"""ABC for calculating the intrinsic quantum yield of photosynthesis.

    This provides an abstract base class for the implementation of alternative
    approaches to calculating the the intrinsic quantum yield of photosynthesis. All
    implementations estimate the :math:`\phi_{0}` following values, which is then stored
    in the ``kphio`` attribute of the resulting class instance.

    The abstract base class requires that implementations of specific approaches defines
    the `calculate_kphio` method. The provides the approach specific calculation of
    ``kphio``  and is automatically called by the ``__init__`` method when a subclass
    instance is created.

    Subclasses must define several class attributes when created:

    .. code:: python

        class QuantumYieldFixed(
            QuantumYieldABC,
            method="method_name",
            requires=["an_environment_variable"],
            array_reference_kphio_ok=True,
        ):

    * The ``method`` argument sets the name of the method, which can then be used to
      select the implemented class from the
      :data:`~pyrealm.pmodel.quantum_yield.QUANTUM_YIELD_CLASS_REGISTRY`.
    * The `requires` argument sets a list of variables that must be present in the
      :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment` to use this
      approach. The core ``tc``, ``vpd``, ``patm`` and ``co2`` variables do not need to
      be included in this list.
    * The ``array_reference_kphio_ok`` argument sets whether the method can accept an
      array of :math:`\phi_0` values or whether a single global reference value should
      be used.

    The definition of the ``_calculate_kphio`` method for subclasses can also provide C3
    and C4 implementations for calculate :math:`\phi_0` - or possibly raise an error for
    one pathway - using the ``use_c4`` attribute.

    Args:
        env: An instance of
            :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment`  providing the
            photosynthetic environment for the model.
        reference_kphio: An optional value to be used instead of the global constant
            value . This is typically a single float but some approaches may support an
            array of values here.
        use_c4: Should the calculation use parameterisation for C4 photosynthesis rather
            than C3 photosynthesis.

    Returns:
        Instances of the abstract base class should not be created - use instances of
        specific subclasses.
    """

    method: str
    """A short method name used to identify the class in
    :data:`~pyrealm.pmodel.quantum_yield.QUANTUM_YIELD_CLASS_REGISTRY`.
    """
    requires: list[str]
    """A list of names of optional attributes of
    :class:`~pyrealm.pmodel.pmodel_environment.PModelEnvironment` that must be populated
    to use a method.
    """
    array_reference_kphio_ok: bool
    """Does the implementation handle arrays inputs to the reference_kphio __init__
    argument."""

    def __init__(
        self,
        env: PModelEnvironment,
        reference_kphio: float | NDArray | None = None,
        use_c4: bool = False,
    ):
        self.env: PModelEnvironment = env
        """The PModelEnvironment containing the photosynthetic environment for the
        model."""
        self.shape: tuple[int, ...] = env.shape
        """The shape of the input environment data."""

        # Set the reference kphio to the class default value if not provided and convert
        # the value to np.array if needed
        if reference_kphio is None:
            reference_kphio = self.env.pmodel_const.maximum_phi0
        if isinstance(reference_kphio, float | int):
            reference_kphio = np.array([reference_kphio])

        # Now check - if the reference_kphio value is a non-scalar array - that array
        # inputs are handled by the kphio method and that the shape matches the shape of
        # the environment.
        if isinstance(reference_kphio, np.ndarray) and reference_kphio.size > 1:
            if self.array_reference_kphio_ok:
                check_input_shapes(self.env.tc, reference_kphio)
            else:
                raise ValueError(
                    f"The {self.method} method for kphio does not support arrays "
                    "of reference kphio values"
                )

        self.reference_kphio: NDArray[np.float64] = reference_kphio
        """The kphio reference value for the method."""
        self.use_c4: bool = use_c4
        """Use a C4 parameterisation if available."""

        # Declare attributes populated by methods. These are typed but not assigned a
        # default value as they must are populated by the subclass specific
        # calculate_kphio method, which is called below to populate the values.
        self.kphio: NDArray[np.float64]
        """The calculated intrinsic quantum yield of photosynthesis."""

        # Run the calculation methods after checking for any required variables
        self._check_requires()
        self._calculate_kphio()

        # Validate that the subclass methods populate the attributes correctly.
        _ = check_input_shapes(env.ca, self.kphio)

    @abstractmethod
    def _calculate_kphio(self) -> None:
        """Calculate the intrinsic quantum yield of photosynthesis."""

    def _check_requires(self) -> None:
        """Check additional required variables are present."""

        for required_var in self.requires:
            if not hasattr(self.env, required_var):
                raise ValueError(
                    f"{self.__class__.__name__} (method {self.method}) requires "
                    f"{required_var} to be provided in the PModelEnvironment."
                )

    def __repr__(self) -> str:
        """Generates a string representation of an QuantumYield instance."""
        return f"{type(self).__name__}(shape={self.shape})"

    def summarize(self, dp: int = 2) -> None:
        """Print QuantumYield summary.

        Prints a summary of the variables calculated within an instance
        of QuantumYield including the mean, range and number of nan values.

        Args:
            dp: The number of decimal places used in rounding summary stats.
        """

        attrs = (("kphio", "-"),)

        summarize_attrs(self, attrs, dp=dp)

    @classmethod
    def __init_subclass__(
        cls,
        method: str,
        requires: list[str],
        array_reference_kphio_ok: bool,
    ) -> None:
        """Initialise a subclass deriving from this ABC."""

        cls.method = method
        cls.requires = requires
        cls.array_reference_kphio_ok = array_reference_kphio_ok
        QUANTUM_YIELD_CLASS_REGISTRY[cls.method] = cls


class QuantumYieldFixed(
    QuantumYieldABC,
    method="fixed",
    requires=[],
    array_reference_kphio_ok=True,
):
    r"""Apply a fixed value for :math:`\phi_0`.

    This implementation applies a fixed value for the quantum yield without any
    environmental variation. It will accept an array of values to allow externally
    estimated values of `phi_0` to be passed to a P Model.
    """

    def _calculate_kphio(self) -> None:
        """Set fixed kphio."""

        self.kphio = self.reference_kphio


class QuantumYieldTemperature(
    QuantumYieldABC,
    method="temperature",
    requires=[],
    array_reference_kphio_ok=False,
):
    r"""Calculate temperature dependent of quantum yield efficiency.

    This implementation calculates temperature dependent quantum yield efficiency, as a
    quadratic function of temperature (:math:`T`).

    .. math::

        \phi(T) = a + b T - c T^2

    The values of :math:`a, b, c` are dependent on whether :math:`\phi_0` is being
    estimated for C3 or C4 photosynthesis. For C3 photosynthesis, the default values use
    the temperature dependence of the maximum quantum yield of photosystem II in
    light-adapted tobacco leaves determined by :cite:t:`Bernacchi:2003dc`. For C4
    photosynthesis, the default values are taken from :cite:t:`cai:2020a`.
    """

    def _calculate_kphio(
        self,
    ) -> None:
        """Calculate kphio."""

        if self.use_c4:
            ftemp = evaluate_horner_polynomial(
                self.env.tc, self.env.pmodel_const.kphio_C4
            )
        else:
            ftemp = evaluate_horner_polynomial(
                self.env.tc, self.env.pmodel_const.kphio_C3
            )

        ftemp = np.clip(ftemp, 0.0, None)
        self.kphio = ftemp * self.reference_kphio


class QuantumYieldSandoval(
    QuantumYieldABC,
    method="sandoval",
    requires=["aridity_index", "mean_growth_temperature"],
    array_reference_kphio_ok=False,
):
    r"""Calculate aridity and mean growth temperature effects on quantum yield.

    This experimental approach implements the method of :cite:t:`sandoval:in_prep`. This
    approach modifies the maximum possible :math:`\phi_0` as a function of the
    climatological aridity index. It then also adjusts the temperature at which the
    highest :math:`\phi_0` can be attained as a function of the mean growth temperature
    for an observation. It then calculates the expected :math:`\phi_0` as a function of
    temperature via a modified Arrhenius relationship.

    The reference kphio for this approach is the theoretical maximum quantum yield,
    defaulting to the ratio of 1/9 in the absence of a Q cycle :cite:`long:1993a`.
    """

    __experimental__: bool = True

    def peak_quantum_yield(
        self, aridity_index: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Calculate the peak quantum yield as a function of the aridity index.

        Args:
            aridity_index: An array of aridity index values.
        """

        # Calculate peak kphio given the aridity index
        m, n = self.env.pmodel_const.sandoval_peak_phio
        return self.reference_kphio / (1 + aridity_index**m) ** n

    def _calculate_kphio(self) -> None:
        """Calculate kphio."""

        # Warn that this is an experimental feature.
        warn_experimental("QuantumYieldSandoval")

        aridity_index = getattr(self.env, "aridity_index")
        mean_growth_temperature = getattr(self.env, "mean_growth_temperature")

        # Calculate enzyme kinetic. This needs to use a copy because the Hd value is
        # modified below.
        coef = self.env.pmodel_const.sandoval_kinetics.copy()

        # Calculate change in activation entropy as a linear function of
        # mean growth temperature, J/mol/K
        delta_entropy = (
            coef["entropy_intercept"] + coef["entropy_slope"] * mean_growth_temperature
        )
        # Calculate de-activation energy J/mol
        Hd = coef["hd"] * delta_entropy

        # Calculate the optimal temperature to be used as the reference temperature in
        # the modified Arrhenius calculation
        Topt = Hd / (
            delta_entropy
            - self.env.core_const.k_R * np.log(coef["ha"] / (Hd - coef["ha"]))
        )
        tk_leaf = self.env.tk
        # Calculate peak kphio given the aridity index
        kphio_peak = self.peak_quantum_yield(aridity_index=aridity_index)

        # Pass the modified Hd back into the calculation of the Arrhenius factor
        coef["hd"] = Hd

        # Calculate the modified Arrhenius factor using the
        f_kphio = calculate_kattge_knorr_arrhenius_factor(
            tk_leaf=tk_leaf,
            tk_ref=Topt,
            tc_growth=mean_growth_temperature,
            coef=coef,
            k_R=self.env.core_const.k_R,
        )

        # Apply the factor and store it.

        # The Sandoval implementation currently includes an additional factor on kphio
        # calculated as (tk_leaf / Topt). This might be the Murphy et al correction, so
        # might be removed, but at present this term is needed to match the R regression
        # test values

        self.kphio = kphio_peak * f_kphio * (tk_leaf / Topt)



"""The isotope_const module TODO."""

from dataclasses import dataclass

from pyrealm.constants import ConstantsClass


@dataclass(frozen=True)
class IsotopesConst(ConstantsClass):
    """Settings for calculate carbon isotope discrimination.

    This data class provides values for underlying constants used in the calculation of
    carbon isotope discrimination from P Model instances.

    * For C4 plants, the coefficients for calculating isotopic discrimination are taken
      from :cite:t:`lavergne:2022a`. The class also provides an alternative, but
      currently implemented, parameterisation using coefficients taken from
      :cite:t:`voncaemmerer:2014a`.
    * For C3 plants, the coefficients for calculating isotopic discrimination are taken
      from :cite:t:`farquhar:1982a`.
    * Post-photosynthetic fractionation values are also provided between leaf organic
      matter and alpha-cellulose :cite:p:`frank:2015a` and bulk wood
      :cite:p:`badeck:2005a`.

    """

    lavergne_delta13_a = 13.95
    """Intercept for isotopic discrimination for C4 plants"""
    lavergne_delta13_b = -17.04
    """Slope for C4 isotopic discrimination for C4 plants"""

    # Farquhar et al. (1982)
    farquhar_a: float = 4.4
    """Intercept for isotopic discrimination for C3 plants"""
    farquhar_b: float = 29
    """Coefficient for simple isotopic discrimination for C3 plants"""
    farquhar_b2: float = 28
    """Coefficient for isotopic discrimination with photorespiration for C3 plants"""
    farquhar_f: float = 12
    """Coefficient for isotopic discrimination with photorespiration for C3 plants"""

    # vonCaemmerer et al. (2014)
    vonCaemmerer_b4: float = -7.4
    """Alternative coefficient for isotopic discrimination for C4 plants"""
    vonCaemmerer_s: float = 1.8
    """Alternative coefficient for isotopic discrimination for C4 plants"""
    vonCaemmerer_phi: float = 0.5
    """Alternative coefficient for isotopic discrimination for C4 plants"""

    frank_postfrac: float = 2.1
    """Post-photosynthetic fractionation between leaf organic matter and
    alpha-cellulose (permil)"""

    badeck_postfrac: float = 1.9
    """Post-photosynthetic fractionation between leaf organic matter and bulk wood
    (permil)"""



"""The :mod:`~pyrealm.core.water` submodule contains core functions for calculating the
density and viscosity of water given the air temperature and atmospheric pressure.
"""  # noqa D210, D415

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import CoreConst
from pyrealm.core.utilities import check_input_shapes, evaluate_horner_polynomial


def calc_density_h2o_chen(
    tc: NDArray[np.float64],
    p: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
) -> NDArray[np.float64]:
    """Calculate the density of water using Chen et al 2008.

    This function calculates the density of water at a given temperature and pressure
    (kg/m^3) following :cite:t:`chen:2008a`.

    Warning:
        The predictions from this function are numerically unstable around -58°C.

    Args:
        tc: Air temperature (°C)
        p: Atmospheric pressure (Pa)
        core_const: Instance of :class:`~pyrealm.constants.core_const.CoreConst`,
            providing the polynomial coefficients for the  :cite:t:`chen:2008a`
            equations.

    Returns:
        Water density in kg/m^3

    Raises:
        ValueError: if the inputs have incompatible shapes.

    Examples:
        >>> round(calc_density_h2o_chen(20, 101325), 3)
        np.float64(998.25)
    """

    # Calculate density at 1 atm (kg/m^3):
    po_coef = core_const.chen_po
    po = evaluate_horner_polynomial(tc, po_coef)

    # Calculate bulk modulus at 1 atm (bar):
    ko_coef = core_const.chen_ko
    ko = evaluate_horner_polynomial(tc, ko_coef)

    # Calculate temperature dependent coefficients:
    ca_coef = core_const.chen_ca
    ca = evaluate_horner_polynomial(tc, ca_coef)

    cb_coef = core_const.chen_cb
    cb = evaluate_horner_polynomial(tc, cb_coef)

    # Convert atmospheric pressure to bar (1 bar = 100000 Pa)
    pbar = (1.0e-5) * p

    pw = ko + ca * pbar + cb * pbar**2.0
    pw /= ko + ca * pbar + cb * pbar**2.0 - pbar
    pw *= (1e3) * po
    return pw


def calc_density_h2o_fisher(
    tc: NDArray[np.float64],
    patm: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
) -> NDArray[np.float64]:
    """Calculate water density.

    Calculates the density of water as a function of temperature and atmospheric
    pressure (kg/m^3), using the Tumlirz Equation and coefficients calculated by
    :cite:t:`Fisher:1975tm`.

    Warning:
        The predictions from this function are unstable around -45°C.

    Args:
        tc: air temperature, °C
        patm: atmospheric pressure, Pa
        core_const: Instance of :class:`~pyrealm.constants.core_const.CoreConst`,
            providing the polynomial coefficients for the :cite:t:`Fisher:1975tm`
            equations.

    Returns:
        Water density in kg/m^3.

    Raises:
        ValueError: if the inputs have incompatible shapes.

    Examples:
        >>> round(calc_density_h2o_fisher(20, 101325), 3)
        np.float64(998.206)
    """

    # Check input shapes, shape not used
    _ = check_input_shapes(tc, patm)

    # Calculate lambda, (bar cm^3)/g:
    lambda_coef = core_const.fisher_dial_lambda
    lambda_val = evaluate_horner_polynomial(tc, lambda_coef)

    # Calculate po, bar
    po_coef = core_const.fisher_dial_Po
    po_val = evaluate_horner_polynomial(tc, po_coef)

    # Calculate vinf, cm^3/g
    vinf_coef = core_const.fisher_dial_Vinf
    vinf_val = evaluate_horner_polynomial(tc, vinf_coef)

    # Convert pressure to bars (1 bar <- 100000 Pa)
    pbar = 1e-5 * patm

    # Calculate the specific volume (cm^3 g^-1):
    spec_vol = vinf_val + lambda_val / (po_val + pbar)

    # Convert to density in kg/m^3
    rho = 1e3 / spec_vol

    return rho


def calc_density_h2o(
    tc: NDArray[np.float64],
    patm: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
    safe: bool = True,
) -> NDArray[np.float64]:
    """Calculate water density.

    Calculates the density of water as a function of temperature and atmospheric
    pressure (in kg/m^3). This function uses either the method provided by
    :cite:t:`Fisher:1975tm` (:func:`~pyrealm.core.water.calc_density_h2o_fisher`) or
    :cite:t:`chen:2008a` (:func:`~pyrealm.core.water.calc_density_h2o_chen`).

    The constants attribute
    :attr:`~pyrealm.constants.core_const.CoreConst.water_density_method` can be used to
    set which of the ``fisher`` or ``chen`` methods is used.

    Args:
        tc: air temperature, °C
        patm: atmospheric pressure, Pa
        core_const: Instance of :class:`~pyrealm.constants.core_const.CoreConst`
        safe: Prevents the function from estimating density below -30°C, where the
            functions are numerically unstable.

    Returns:
        Water density in kg/m^3.

    Raises:
        ValueError: if ``tc`` contains values below -30°C and ``safe`` is True, or if
            the inputs have incompatible shapes.

    Examples:
        >>> round(calc_density_h2o(20, 101325), 3)
        np.float64(998.206)
    """

    # Safe guard against instability in functions at low temperature.
    if safe and np.nanmin(tc) < np.array([-30]):
        raise ValueError(
            "Water density calculations below about -30°C are "
            "unstable. See argument safe to calc_density_h2o"
        )

    # Check input shapes, shape not used
    _ = check_input_shapes(tc, patm)

    if core_const.water_density_method == "fisher":
        return calc_density_h2o_fisher(tc, patm, core_const)

    if core_const.water_density_method == "chen":
        return calc_density_h2o_chen(tc, patm, core_const)

    raise ValueError("Unknown method provided to calc_density_h2o")


def calc_viscosity_h2o(
    tc: NDArray[np.float64],
    patm: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
    simple: bool = False,
) -> NDArray[np.float64]:
    r"""Calculate the viscosity of water.

    Calculates the viscosity of water (:math:`\eta`) as a function of temperature and
    atmospheric pressure :cite:p:`Huber:2009fy`.

    Args:
        tc: air temperature (°C)
        patm: atmospheric pressure (Pa)
        core_const: Instance of :class:`~pyrealm.constants.core_const.CoreConst`
        simple: Use the simple formulation.

    Returns:
        A float giving the viscosity of water (mu, Pa s)

    Examples:
        >>> # Density of water at 20 degrees C and standard atmospheric pressure:
        >>> round(calc_viscosity_h2o(20, 101325), 7)
        np.float64(0.0010016)
    """

    # Check inputs, return shape not used
    _ = check_input_shapes(tc, patm)

    if simple or core_const.simple_viscosity:
        # The reference for this is unknown, but is used in some implementations
        # so is included here to allow intercomparison.
        return np.exp(-3.719 + 580 / ((tc + 273) - 138))

    # Get the density of water, kg/m^3
    rho = calc_density_h2o(tc, patm, core_const=core_const)

    # Calculate dimensionless parameters:
    tbar = (tc + core_const.k_CtoK) / core_const.huber_tk_ast
    rbar = rho / core_const.huber_rho_ast

    # Calculate mu0 (Eq. 11 & Table 2, Huber et al., 2009):
    mu0 = core_const.huber_H_i[0] + core_const.huber_H_i[1] / tbar
    mu0 += core_const.huber_H_i[2] / (tbar * tbar)
    mu0 += core_const.huber_H_i[3] / (tbar * tbar * tbar)
    mu0 = (1e2 * np.sqrt(tbar)) / mu0

    # Calculate mu1 (Eq. 12 & Table 3, Huber et al., 2009):
    ctbar = (1.0 / tbar) - 1.0
    mu1 = 0.0

    # Iterate over the rows of the H_ij core_constants matrix
    for row_idx in np.arange(core_const.huber_H_ij.shape[1]):
        cf1 = ctbar**row_idx
        cf2 = 0.0
        for col_idx in np.arange(core_const.huber_H_ij.shape[0]):
            cf2 += core_const.huber_H_ij[col_idx, row_idx] * (rbar - 1.0) ** col_idx
        mu1 += cf1 * cf2

    mu1 = np.exp(rbar * mu1)

    # Calculate mu_bar (Eq. 2, Huber et al., 2009), assumes mu2 = 1
    mu_bar = mu0 * mu1

    # Calculate mu (Eq. 1, Huber et al., 2009)
    return mu_bar * core_const.huber_mu_ast  # Pa s


def calc_viscosity_h2o_matrix(
    tc: NDArray[np.float64],
    patm: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
    simple: bool = False,
) -> NDArray[np.float64]:
    r"""Calculate the viscosity of water.

    Calculates the viscosity of water (:math:`\eta`) as a function of temperature and
    atmospheric pressure :cite:p:`Huber:2009fy`.

    Args:
        tc: air temperature (°C)
        patm: atmospheric pressure (Pa)
        core_const: Instance of :class:`~pyrealm.constants.core_const.CoreConst`
        simple: Use the simple formulation.

    Returns:
        A float giving the viscosity of water (mu, Pa s)

    Examples:
        >>> # Viscosity of water at 20 degrees C and standard atmospheric pressure:
        >>> round(calc_viscosity_h2o(20, 101325), 7)
        np.float64(0.0010016)
    """

    # Check inputs, return shape not used
    _ = check_input_shapes(tc, patm)

    if simple or core_const.simple_viscosity:
        # The reference for this is unknown, but is used in some implementations
        # so is included here to allow intercomparison.
        return np.exp(-3.719 + 580 / ((tc + 273) - 138))

    # Get the density of water, kg/m^3
    rho = calc_density_h2o(tc, patm, core_const=core_const)

    # Calculate dimensionless parameters:
    tbar = (tc + core_const.k_CtoK) / core_const.huber_tk_ast
    rbar = rho / core_const.huber_rho_ast

    # Calculate mu0 (Eq. 11 & Table 2, Huber et al., 2009):
    tbar_pow = np.power.outer(tbar, np.arange(0, 4))
    mu0 = (1e2 * np.sqrt(tbar)) / np.sum(
        np.array(core_const.huber_H_i) / tbar_pow, axis=-1
    )

    # Calculate mu1 (Eq. 12 & Table 3, Huber et al., 2009):
    h_array = np.array(core_const.huber_H_ij)
    ctbar = (1.0 / tbar) - 1.0
    row_j, _ = np.indices(h_array.shape)
    mu1 = h_array * np.power.outer(rbar - 1.0, row_j)
    mu1 = np.power.outer(ctbar, np.arange(0, 6)) * np.sum(mu1, axis=(-2))
    mu1 = np.exp(rbar * mu1.sum(axis=-1))

    # Calculate mu_bar (Eq. 2, Huber et al., 2009), assumes mu2 = 1
    mu_bar = mu0 * mu1

    # Calculate mu (Eq. 1, Huber et al., 2009)
    return mu_bar * core_const.huber_mu_ast  # Pa s


def convert_water_mm_to_moles(
    water_mm: NDArray[np.float64],
    tc: NDArray[np.float64],
    patm: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
) -> NDArray[np.float64]:
    """Convert water in mm per square meter to moles.

    This function converts water volumes expressed as mm per m2 into a number of moles
    of water. It accounts for the changing density of water with temperature and
    pressure.

    Args:
        water_mm: Water volume in mm per square meter
        tc: air temperature (°C)
        patm: atmospheric pressure (Pa)
        core_const: Instance of :class:`~pyrealm.constants.core_const.CoreConst`

    Returns:
        Moles of water (-)

    Examples:
        >>> # At 0°C and 101325 Pa, one mole of water is ~18 g (18 cm3, 0.018 mm m-2).
        >>> # So, 1 mm m2 = 1 / 0.018 = ~55 moles.
        >>> round(convert_water_mm_to_moles(water_mm=1, tc=0, patm=101325), 3)
        np.float64(55.508)
    """

    # Check inputs, return shape not used
    _ = check_input_shapes(water_mm, tc, patm)

    # 1 mm per square meter is 1 litre, so convert to mL and then to moles
    return (
        water_mm
        * 1000
        / calculate_water_molar_volume(tc=tc, patm=patm, core_const=core_const)
    )


def convert_water_moles_to_mm(
    water_moles: NDArray[np.float64],
    tc: NDArray[np.float64],
    patm: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
) -> NDArray[np.float64]:
    """Convert water in moles to mm per square meter.

    This function converts water volumes expressed as moles into mm per m2. It accounts
    for the changing density of water with temperature and pressure.

    Args:
        water_moles: Water volume in moles
        tc: air temperature (°C)
        patm: atmospheric pressure (Pa)
        core_const: Instance of :class:`~pyrealm.constants.core_const.CoreConst`

    Returns:
        Water volume in mm per m2

    Examples:
        >>> # At 0°C and 101325 Pa, one mole of water is ~18 g (18 cm3, 0.018 mm m-2).
        >>> # So, 1 mol = 0.018 mm
        >>> round(convert_water_moles_to_mm(water_moles=1, tc=0, patm=101325), 3)
        np.float64(0.018)
    """

    # Check inputs, return shape not used
    _ = check_input_shapes(water_moles, tc, patm)

    # 1 mm per square meter is 1 litre, so convert to mL and then to moles
    return (
        water_moles
        * calculate_water_molar_volume(tc=tc, patm=patm, core_const=core_const)
    ) / 1000


def calculate_water_molar_volume(
    tc: NDArray[np.float64],
    patm: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
) -> NDArray[np.floating]:
    """Calculate the volume of a mole of water at a given temperature and pressure.

    Args:
        tc: air temperature (°C)
        patm: atmospheric pressure (Pa)
        core_const: Instance of :class:`~pyrealm.constants.core_const.CoreConst`

    Returns:
        Water molar volume in mol cm-3, or equivalently mol/mL

    Examples:
        >>> # A mole of water at standard temperature and pressure occupies ~18 cm3.
        >>> round(calculate_water_molar_volume(0, 101235), 3)
        np.float64(18.015)
    """
    # Calculate density at given temperature and pressure in g/cm3
    water_density = calc_density_h2o(tc=tc, patm=patm, core_const=core_const) / 1000
    # Hence molar volume as mol/cm3 or equivalently mol/mL
    return core_const.k_water_molmass / water_density



"""The phenology_const module TODO."""

from dataclasses import dataclass

from pyrealm.constants import ConstantsClass


@dataclass(frozen=True)
class PhenologyConst(ConstantsClass):
    r"""Model constants for the phenology module class.

    This data class holds constants needed for the FaparLimitation and other classes
    in the phenology module.
    """

    z: float = 12.227
    """z accounts for the costs of building and maintaining leaves and the total
       below-ground allocation required to support the nutrient demand of those
       leaves. [mol m^{-2} year^{-1}]."""

    k: float = 0.5
    """Light extinction coefficient."""

    f0_coefficients = (0.65, 0.604169, 1.9)
    """Coefficients to calculate f_0, the ratio of annual total transpiration of annual 
    total precipitation."""



"""The competition_const module TODO."""

from dataclasses import dataclass

from pyrealm.constants import ConstantsClass


@dataclass(frozen=True)
class C3C4Const(ConstantsClass):
    r"""Model constants for the C3C4Competition class.

    This data class holds statistically estimated coefficients used to calculate the
    fraction of C4 plants based on the relative GPP of C3 and C4 plants for given
    conditions and estimated treecover :cite:p:`lavergne:2020a`.
    """

    # Non-linear regression of fraction C4 plants from proportion GPP advantage
    # of C4 over C3 plants
    adv_to_frac_k: float = 6.63
    """Coefficient k of C4 fraction estimation"""
    adv_to_frac_q: float = 0.16
    """Coefficient 1 of C4 fraction estimation"""

    # Conversion parameters to estimate tree cover from  C3 GPP
    gpp_to_tc_a: float = 15.60
    """Coefficient a of tree cover estimation"""
    gpp_to_tc_b: float = 1.41
    """Coefficient b of tree cover estimation"""
    gpp_to_tc_c: float = -7.72
    """Coefficient c of tree cover estimation"""
    c3_forest_closure_gpp: float = 2.8
    """GPP at which forest canopy closure occurs"""



"""The demography module contains functions and classes for modelling the demography of
tree communities, including the definition of plant functional types, size structured
plant cohorts, community canopy models and the allocation of gross primary productivity
into respiration, turnover and growth.
"""  # noqa: D205



"""Many of the functions and classes in `pyrealm` have an underlying set of
parameters that will typically be held constant. This includes true universal constants,
such as the molar mass of Carbon, but also a large number of calculated parameters from
the literature.

These underlying parameters are not hardcoded, but rather dataclasses for constants  are
defined for different models within `pyrealm`. The defaults in these dataclasses can be
altered and can also be loaded and saved to configuration files. The
:class:`~pyrealm.constants.base.ConstantsClass` base class provides the basic load and
save methods and then individual subclasses define the constants and default values for
each class.

This implementation has the following desired features:

1. Ability to use a obj.attr notation rather than obj['attr'].
2. Ability to freeze values to avoid them being edited in use - distinctly paranoid!
3. Ability to set a default mapping with default values.
4. Typing to set expected types on default values.
5. Simple export/import methods to go to from dict / JSON.
6. Is a class, to allow __repr__ and other methods.
"""  # noqa D210, D415

import json
from dataclasses import asdict, dataclass

from dacite import from_dict


@dataclass(frozen=True)
class ConstantsClass:
    """Base class for model constants.

    This base class provides a consistent interface for creating and exporting data from
    model constant classes. It defines methods to create instances from a dictionary of
    values and from a JSON file of values. It also defines methods to export an existing
    instance of a constants class to a dictionary or a JSON file.

    All model constant classes use the :func:`~dataclasses.dataclass` decorator as the
    basic structure. All fields in these classes are defined with default values, so
    instances can be defined using partial dictionaries of alternative values.
    """

    def to_dict(self) -> dict:
        """Return a parameter dictionary.

        This method returns the attributes and values used in an instance of a parameter
        class object as a dictionary.

        Returns:
            A dictionary
        """
        return asdict(self)

    def to_json(self, filename: str) -> None:
        """Export a parameter set to JSON.

        This method writes the attributes and values used in an instance of a
        parameter class object to a JSON file.

        Args:
            filename: A path to the JSON file to be created.

        Returns:
            None
        """
        with open(filename, "w") as outfile:
            json.dump(self.to_dict(), outfile, indent=4)

    @classmethod
    def from_dict(cls, data: dict) -> "ConstantsClass":
        """Create a ConstantsClass subclass instance from a dictionary.

        Generates a :class:`~pyrealm.constants.base.ConstantsClass` subclass instance
        using the data provided in a dictionary to override default values

        Args:
          data: A dictionary, keyed by parameter class attribute names, providing values
            to use in a new instance.

        Returns:
            An instance of the subclass.
        """
        return from_dict(cls, data)

    @classmethod
    def from_json(cls, filename: str) -> "ConstantsClass":
        """Create a ParamClass instance from a JSON file.

        Generates a :class:`~pyrealm.constants.base.ConstantsClass` subclass instance
        using the data provided in a  JSON file.

        Args:
          filename: The path to a JSON formatted file containing a set of values keyed
            by parameter class attribute names to use in a new instance.

        Returns:
            An instance of the parameter class.
        """
        with open(filename) as infile:
            json_data = json.load(infile)

        return cls.from_dict(json_data)



"""The :mod:`~pyrealm.pmodel.functions` submodule contains the main standalone functions
used for calculating the photosynthetic behaviour of plants. The documentation describes
the key equations used in each function.
"""  # noqa D210, D415

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import CoreConst, PModelConst
from pyrealm.core.utilities import check_input_shapes
from pyrealm.core.water import calc_viscosity_h2o


def calculate_simple_arrhenius_factor(
    tk: NDArray[np.float64],
    tk_ref: float,
    ha: float,
    k_R: float = CoreConst().k_R,
) -> NDArray[np.float64]:
    r"""Calculate an Arrhenius scaling factor using activation energy.

    Calculates the temperature-scaling factor :math:`f` for enzyme kinetics following
    a simple Arrhenius response governed solely by the activation energy for an enzyme
    (``ha``, :math:`H_a`). The rate is given for a temperature :math:`T` relative to a
    reference temperature :math:T_0`, both given in Kelvin.

    Arrhenius kinetics are described as:

    .. math::

        x(T) = \exp(c - H_a / (T R))

    The temperature-correction function :math:`f(T, H_a)` is:

      .. math::
        :nowrap:

        \[
            \begin{align*}
                f &= \frac{x(T)}{x(T_0)} \\
                  &= \exp \left( \frac{ H_a (T - T_0)}{T_0 R T}\right)
                        \text{, or equivalently}\\
                  &= \exp \left( \frac{ H_a}{R} \cdot
                        \left(\frac{1}{T_0} - \frac{1}{T}\right)\right)
            \end{align*}
        \]

    Args:
        tk: Temperature (K)
        tk_ref: The reference temperature for the reaction (K).
        ha: Activation energy (in :math:`J \text{mol}^{-1}`)
        k_R: The universal gas constant, defaulting to the value from
            attr:`~pyrealm.constants.core_const.CoreConst.k_R`. 
    
    Examples:
        >>> # Percentage rate change from 25 to 10 degrees Celsius
        >>> at_10C = calculate_simple_arrhenius_factor(
        ...     np.array([283.15]) , 298.15, 100000
        ... )
        >>> np.round((1.0 - at_10C) * 100, 4)
        array([88.1991])
    """

    return np.exp(ha * (tk - tk_ref) / (tk_ref * k_R * tk))


def calculate_kattge_knorr_arrhenius_factor(
    tk_leaf: NDArray[np.float64],
    tk_ref: float,
    tc_growth: NDArray[np.float64],
    coef: dict[str, float],
    k_R: float = CoreConst().k_R,
) -> NDArray[np.float64]:
    r"""Calculate an Arrhenius factor following :cite:t:`Kattge:2007db`.

    This implements a "peaked" version of the Arrhenius relationship, describing a
    decline in reaction rates at higher temperatures. In addition to the activation
    energy (see :meth:`~pyrealm.pmodel.functions.calculate_simple_arrhenius_factor`),
    this implementation adds an entropy term and the deactivation energy of the enzyme
    system. The rate is given for a given instantaneous temperature :math:`T` relative
    to a reference temperature :math:T_0`, both given in Kelvin, but the entropy is
    calculated using a separate estimate of the growth temperature for a plant,
    expressed in °C.


    .. math::
        :nowrap:

        \[
            \begin{align*}

                f  &= \exp \left( \frac{ H_a (T - T_0)}{T_0 R T}\right)
                      \left(
                        \frac{1 + \exp \left( \frac{T_0 \Delta S - H_d }{ R T_0}\right)}
                             {1 + \exp \left( \frac{T \Delta S - H_d}{R T} \right)}
                      \right)
                      \left(\frac{T}{T_0}\right)
            \end{align*}

            \text{where,}

            \Delta S = a + b * t_g

        \]

    The coefficients dictionary must provide entries for:

    * ha: The activation energy of the enzyme (:math:`H_a`)
    * hd: The deactivation energy of the enzyme (:math:`H_d`)
    * entropy_intercept: The intercept of the entropy relationship (:math:`a`)
    * entropy_slope: The slope of the entropy relationship (:math:`b`)

    Args:
        tk_leaf: The instantaneous temperature in Kelvin (K) at which to calculate the
            factor (:math:`T`)
        tk_ref: The reference temperature in Kelvin for the process (:math:`T_0`)
        tc_growth: The growth temperature of the plants in °C (:math:`t_g`)
        coef: A dictionary providing values of the coefficients ``ha``,
            ``hd``, ``entropy_intercept`` and ``entropy_slope``.
        k_R: The universal gas constant, defaulting to the value from
            attr:`~pyrealm.constants.core_const.CoreConst.k_R`.

    Returns:
        Values for :math:`f`

    Examples:
        >>> # Calculate the factor for the relative rate of Vcmax at 10 °C (283.15K)
        >>> # compared to the rate at the reference temperature of 25°C (298.15K).
        >>> from pyrealm.constants import PModelConst
        >>> pmodel_const = PModelConst()
        >>> # Get enzyme kinetics parameters
        >>> coef = pmodel_const.arrhenius_vcmax['kattge_knorr']
        >>> # Calculate the arrhenius factor
        >>> val = calculate_kattge_knorr_arrhenius_factor(
        ...     tk_leaf= np.array([283.15]),
        ...     tc_growth = 10,
        ...     tk_ref=298.15,
        ...     coef=coef,
        ... )
        >>> np.round(val, 4)
        array([0.261])
    """

    # Calculate entropy as a function of temperature _in °C_
    entropy = coef["entropy_intercept"] + coef["entropy_slope"] * tc_growth

    # Calculate Arrhenius components
    fva = calculate_simple_arrhenius_factor(
        tk=tk_leaf, ha=coef["ha"], tk_ref=tk_ref, k_R=k_R
    )

    fvb = (1 + np.exp((tk_ref * entropy - coef["hd"]) / (k_R * tk_ref))) / (
        1 + np.exp((tk_leaf * entropy - coef["hd"]) / (k_R * tk_leaf))
    )

    return fva * fvb


def calc_ftemp_inst_rd(
    tc: NDArray[np.float64],
    tc_ref: float = PModelConst().tc_ref,
    coef: tuple[float, float] = PModelConst().heskel_rd,
) -> NDArray[np.float64]:
    r"""Calculate temperature scaling of dark respiration.

    Calculates the temperature-scaling factor for dark respiration at a given
    temperature (``tc``, :math:`T` in °C), relative to the standard reference
    temperature :math:`T_o`, given the parameterisation in :cite:t:`Heskel:2016fg`.

    .. math::

            fr = \exp( b (T_o - T) -  c ( T_o^2 - T^2 ))

    Args:
        tc: Temperature (:math:`T`, °C)
        tc_ref: standard reference temperature for photosynthetic processes
            (:math:`T_o`,°C)
        coef: A two tuple of floats providing the linear and quadratic coefficients
            (:math:`b` and :math:`c`)

    Examples:
        >>> # Relative instantaneous change in Rd going from 10 to 25 degrees
        >>> pmod_consts = PModelConst()
        >>> (
        ...     calc_ftemp_inst_rd(
        ...         tc=25, tc_ref=pmod_consts.tc_ref, coef=pmod_consts.heskel_rd
        ...     )
        ...     / calc_ftemp_inst_rd(
        ...         tc=10, tc_ref=pmod_consts.tc_ref, coef=pmod_consts.heskel_rd
        ...     )
        ...     - 1
        ... ).round(4)
        np.float64(2.5096)
    """

    return np.exp(coef[0] * (tc - tc_ref) - coef[1] * (tc**2 - tc_ref**2))


def calc_gammastar(
    tk: NDArray[np.float64],
    patm: NDArray[np.float64],
    tk_ref: float = PModelConst().tk_ref,
    k_Po: float = CoreConst().k_Po,
    k_R: float = CoreConst().k_R,
    coef: dict[str, float] = PModelConst().bernacchi_gs,
) -> NDArray[np.float64]:
    r"""Calculate the photorespiratory CO2 compensation point.

    Calculates the photorespiratory **CO2 compensation point** in absence of dark
    respiration (:math:`\Gamma^{*}`, :cite:alp:`Farquhar:1980ft`) as:

    .. math::

        \Gamma^{*} = \Gamma^{*}_{0} \cdot \frac{p}{p_0} \cdot f(T, H_a)

    where :math:`f(T, H_a)` modifies the activation energy to the the local temperature
    in Kelvin following the Arrhenius-type temperature response function (see
    :meth:`~pyrealm.pmodel.functions.calculate_simple_arrhenius_factor`). By default,
    estimates of  :math:`\Gamma^{*}_{0}` and :math:`H_a` are taken from
    :cite:t:`Bernacchi:2001kg` (see
    :attr:`PModelConst.bernacchi_gs<pyrealm.constants.pmodel_const.PModelConst.bernacchi_gs>`)

    Args:
        tk: Temperature relevant for photosynthesis in Kelvin(:math:`T`, K)
        patm: Atmospheric pressure (:math:`p`, Pascals)
        tk_ref: The reference temperature of the coefficients in Kelvin.
        k_Po: The standard atmospheric pressure, defaulting to the value
            from :attr:`~pyrealm.constants.core_const.CoreConst.k_Po`.
        k_R: The universal gas constant, defaulting to the value from
            :attr:`~pyrealm.constants.core_const.CoreConst.k_R`.
        coef: A dictionary providing the enzyme kinetic coefficients for the reaction
            (``dha`` and ``gs25_0``).

    Returns:
        A float value or values for :math:`\Gamma^{*}` (in Pa)

    Examples:
        >>> # CO2 compensation point at 20 °C  (293.15 K) and standard presssure
        >>> calc_gammastar(np.array([293.15]), np.array([101325])).round(5)
        array([3.33925])
    """

    # check inputs, return shape not used
    _ = check_input_shapes(tk, patm)

    return (
        coef["gs25_0"]
        * patm
        / k_Po
        * calculate_simple_arrhenius_factor(
            tk=tk, tk_ref=tk_ref, ha=coef["dha"], k_R=k_R
        )
    )


def calc_ns_star(
    tc: NDArray[np.float64],
    patm: NDArray[np.float64],
    core_const: CoreConst = CoreConst(),
) -> NDArray[np.float64]:
    r"""Calculate the relative viscosity of water.

    Calculates the relative viscosity of water (:math:`\eta^*`), given the standard
    temperature and pressure, using :func:`~pyrealm.core.water.calc_viscosity_h2o`
    (:math:`v(t,p)`) as:

    .. math::

        \eta^* = \frac{v(t,p)}{v(t_0,p_0)}

    Args:
        tc: Temperature, relevant for photosynthesis (:math:`T`, °C)
        patm: Atmospheric pressure (:math:`p`, Pa)
        core_const: Instance of :class:`~pyrealm.constants.core_const.CoreConst`.

    PModel Parameters:
        To: standard temperature (:math:`t0`, ``k_To``)
        Po: standard pressure (:math:`p_0`, ``k_Po``)

    Returns:
        A numeric value for :math:`\eta^*` (a unitless ratio)

    Examples:
        >>> # Relative viscosity at 20 degrees Celsius and standard
        >>> # atmosphere (in Pa):
        >>> round(calc_ns_star(20, 101325), 5)
        np.float64(1.12536)
    """

    visc_env = calc_viscosity_h2o(tc, patm, core_const=core_const)
    visc_std = calc_viscosity_h2o(
        np.array(core_const.k_To) - np.array(core_const.k_CtoK),
        np.array(core_const.k_Po),
        core_const=core_const,
    )

    return visc_env / visc_std


def calc_kmm(
    tk: NDArray[np.float64],
    patm: NDArray[np.float64],
    tk_ref: float = PModelConst().tk_ref,
    k_co: float = CoreConst().k_co,
    k_R: float = CoreConst().k_R,
    coef: dict[str, float] = PModelConst().bernacchi_kmm,
) -> NDArray[np.float64]:
    r"""Calculate the Michaelis Menten coefficient of Rubisco-limited assimilation.

    Calculates the Michaelis Menten coefficient of Rubisco-limited assimilation
    (:math:`K`, :cite:alp:`Farquhar:1980ft`) as a function of temperature (:math:`T,
    Kelvin) and atmospheric pressure (:math:`p`, Pa) as:

      .. math:: K = K_c ( 1 + p_{\ce{O2}} / K_o),

    where, :math:`p_{\ce{O2}} = 0.209476 \cdot p` is the partial pressure of oxygen.
    :math:`f(T, H_a)` is the simple Arrhenius temperature response of activation
    energies (see :meth:`~pyrealm.pmodel.functions.calculate_simple_arrhenius_factor`)
    used to correct Michalis constants at standard temperature for both :math:`\ce{CO2}`
    and :math:`\ce{O2}` to the local temperature. The default values for the enzyme
    coefficients are taken from Table 1 of :cite:t:`Bernacchi:2001kg` (see
    attr:`PModelConst.bernacchi_kmm<pyrealm.constants.pmodel_const.PModelConst.bernacchi_kmm>`)

      .. math::
        :nowrap:

        \[
            \begin{align*}
                K_c &= K_{c25} \cdot f(T, H_{kc})\\ K_o &= K_{o25} \cdot f(T, H_{ko})
            \end{align*}
        \]

    Args:
        tk: Temperature relevant for photosynthesis in Kelvin (:math:`T`, K)
        patm: Atmospheric pressure (:math:`p`, Pa)
        tk_ref: The reference temperature of the coefficients in Kelvin.
        k_co: The partial pressure of :math:`\ce{O2}` at standard pressure, defaulting
            to the values from  attr:`~pyrealm.constants.core_const.CoreConst.k_co`.
        k_R: The universal gas constant, defaulting to the value from
            attr:`~pyrealm.constants.core_const.CoreConst.k_R`.
        coef: A dictionary providing the enzyme kinetic coefficients for the reaction
            (``kc25``, ``ko25``, ``dhac``, ``dhao``).

    Returns:
        A numeric value for :math:`K` (in Pa)

    Examples:
        >>> # Michaelis-Menten coefficient at 20°C (293.15K) and standard pressure (Pa)
        >>> calc_kmm(np.array([293.15]), np.array([101325])).round(5)
        array([46.09928])
    """

    # Check inputs, return shape not used
    _ = check_input_shapes(tk, patm)

    kc = coef["kc25"] * calculate_simple_arrhenius_factor(
        tk=tk, tk_ref=tk_ref, ha=coef["dhac"], k_R=k_R
    )

    ko = coef["ko25"] * calculate_simple_arrhenius_factor(
        tk=tk, tk_ref=tk_ref, ha=coef["dhao"], k_R=k_R
    )

    # O2 partial pressure
    po = k_co * 1e-6 * patm

    return kc * (1.0 + po / ko)


def calc_soilmstress_stocker(
    soilm: NDArray[np.float64],
    meanalpha: NDArray[np.float64] = np.array(1.0),
    coef: dict[str, float] = PModelConst().soilmstress_stocker,
) -> NDArray[np.float64]:
    r"""Calculate Stocker's empirical soil moisture stress factor.

    This function calculates a penalty factor :math:`\beta(\theta)` for well-watered GPP
    estimates as an empirically derived stress factor :cite:p:`Stocker:2020dh`. The
    factor is calculated as a function of relative soil moisture (:math:`m_s`, fraction
    of field capacity) and average aridity, quantified by the local annual mean ratio of
    actual over potential evapotranspiration (:math:`\bar{\alpha}`).

    The value of :math:`\beta` is defined relative to two soil moisture thresholds
    (:math:`\theta_0, \theta^{*}`) as:

      .. math::
        :nowrap:

        \[
            \beta =
                \begin{cases}
                    q(m_s - \theta^{*})^2 + 1,  & \theta_0 < m_s <= \theta^{*} \\
                    1, &  \theta^{*} < m_s,
                \end{cases}
        \]

    where :math:`q` is an aridity sensitivity parameter setting the stress factor at
    :math:`\theta_0`:

    .. math:: q=(1 - (a + b \bar{\alpha}))/(\theta^{*} - \theta_{0})^2

    .. IMPORTANT::

        The default parameterisation of this water stress penalty (:math:`a=0`,
        :math:`b=0.7330`)  was estimated from empirical data using the **standard form
        of the PModel**  (:class:`~pyrealm.pmodel.pmodel.PModel` in ``pyrealm``).

        These parameters were then further calibrated against empirical data by tuning
        the quantum yield of photosynthesis. This tuning aimed to capture include
        incomplete leaf absorption in the realised value of :math:`\phi_0`, and
        :cite:t:`Stocker:2020dh` argue that, within their model representation,
        :math:`\phi_0` should be treated as a parameter representing canopy-scale
        effective quantum yield. To duplicate the model settings used with this soil
        moisture correction in Table 1 of :cite:t:`Stocker:2020dh`, use the following
        code:

        .. code-block:: python

            # The 'BRC' model setup
            PModel(
                ...
                method_kphio="temperature",
                method_arrhenius="simple",
                method_jmaxlim="wang17",
                method_optchi="prentice14",
                reference_kphio=0.081785,
            )

            # The 'ORG' model setup
            PModel(
                ...
                method_kphio="fixed",
                method_arrhenius="simple",
                method_jmaxlim="wang17",
                method_optchi="prentice14",
                reference_kphio=0.049977,
            )

    The :mod:`pyrealm.pmodel` module treats this factor purely as a penalty that can be
    applied after the estimation of GPP. In contrast, the `rpmodel` implementation uses
    the penalised GPP to back-calculate realistic :math:`J_{max}` and :math:`V_{cmax}`
    values that would give rise to the penalised GPP.

    Args:
        soilm: Relative soil moisture as a fraction of field capacity
            (unitless). Defaults to 1.0 (no soil moisture stress).
        meanalpha: Local annual mean ratio of actual over potential
            evapotranspiration, measure for average aridity. Defaults to 1.0.
        coef: A dictionary providing values of the coefficients ``theta0``,
            ``thetastar``, ``a`` and ``b``, defaulting to the values from 
            attr:`~pyrealm.constants.pmodel_const.PModelConst.soilmstress_stocker`.

    Returns:
        A numeric value or values for :math:`\beta`

    Examples:
        >>> # Proportion of well-watered GPP available at soil moisture of 0.2
        >>> calc_soilmstress_stocker(np.array([0.2])).round(5)
        array([0.88133])
    """

    # TODO - move soilm params into standalone param class for this function -
    #        keep the PModelConst cleaner?

    # Check inputs, return shape not used
    _ = check_input_shapes(soilm, meanalpha)

    # Calculate outstress
    y0 = coef["a"] + coef["b"] * meanalpha
    beta = (1.0 - y0) / (coef["theta0"] - coef["thetastar"]) ** 2
    outstress = 1.0 - beta * (soilm - coef["thetastar"]) ** 2

    # Filter wrt to thetastar
    outstress = np.where(soilm <= coef["thetastar"], outstress, 1.0)

    # Clip
    outstress = np.clip(outstress, 0.0, 1.0)

    return outstress


def calc_soilmstress_mengoli(
    soilm: NDArray[np.float64] = np.array(1.0),
    aridity_index: NDArray[np.float64] = np.array(1.0),
    coef: dict[str, float] = PModelConst().soilmstress_mengoli,
) -> NDArray[np.float64]:
    r"""Calculate the Mengoli et al. empirical soil moisture stress factor.

    This function calculates a penalty factor :math:`\beta(\theta)` for well-watered GPP
    estimates as an empirically derived stress factor :cite:p:`mengoli:2023a`. The
    factor is calculated from relative soil moisture as a fraction of field capacity
    (:math:`\theta`) and the long-run climatological aridity index for a site
    (:math:`\textrm{AI}`), calculated as (total PET)/(total precipitation) for a
    suitable time period.

    The factor is calculated using two constrained power functions for the maximal level
    (:math:`y`) of productivity and the threshold (:math:`psi`) at which that maximal
    level is reached.

      .. math::
        :nowrap:

        \[
            \begin{align*}
            y &= \min( a  \textrm{AI} ^ {b}, 1)\\
            \psi &= \min( a  \textrm{AI} ^ {b}, 1)\\
            \beta(\theta) &=
                \begin{cases}
                    y, & \theta \ge \psi \\
                    \dfrac{y}{\psi} \theta, & \theta \lt \psi \\
                \end{cases}\\
            \end{align*}
        \]

    .. IMPORTANT::

        The parameterisation of this water stress penalty was estimated from empirical
        data using the **subdaily form of the PModel**
        (:class:`~pyrealm.pmodel.pmodel.SubdailyPModel` in ``pyrealm``) with
        temperature dependence of the standard maximum quantum yield of photosynthesis
        (``phi_0``, :math:`\phi_0=1/8`). 
        
        There are minor differences in the implementation of the Subdaily P Model in
        ``pyrealm`` from that used to calibrate this function in
        :cite:p:`mengoli:2023a`. To get the closest match when applying this
        soil moisture correction, use the following settings:

        .. code-block:: python

            SubdailyPModel(
                ...
                method_kphio="temperature",
                method_arrhenius="simple",
                method_jmaxlim="wang17",
                method_optchi="prentice14",
                reference_kphio=1/8,
            )

        The ``reference_kphio=1/8`` value here is in fact the default value used when
        ``method_kphio="temperature"`` but is restated here for clarity.

    Args:
        soilm: Relative soil moisture (unitless).
        aridity_index: The climatological aridity index.
        coef: A dictionary providing values of the coefficients ``y_a``, ``y_b``,
            ``psi_a`` and ``psi_b``, defaulting to the values from 
            attr:`~pyrealm.constants.pmodel_const.PModelConst.soilmstress_mengoli`.

    Returns:
        A numeric value or values for :math:`f(\theta)`

    Examples:
        >>> import numpy as np
        >>> # Proportion of well-watered GPP available with soil moisture and aridity
        >>> # index values of 0.6
        >>> calc_soilmstress_mengoli(np.array([0.6]), np.array([0.6])).round(5)
        array([0.78023])
    """

    # TODO - move soilm params into standalone param class for this function -
    #        keep the PModelConst cleaner?

    # Check inputs, return shape not used
    _ = check_input_shapes(soilm, aridity_index)

    # Calculate maximal level and threshold
    y = np.minimum(
        coef["y_a"] * np.power(aridity_index, coef["y_b"]),
        1,
    )

    psi = np.minimum(
        coef["psi_a"] * np.power(aridity_index, coef["psi_b"]),
        1,
    )

    # Return factor
    return np.where(soilm >= psi, y, (y / psi) * soilm)


def calc_co2_to_ca(
    co2: NDArray[np.float64], patm: NDArray[np.float64]
) -> NDArray[np.float64]:
    r"""Convert :math:`\ce{CO2}` ppm to Pa.

    Converts ambient :math:`\ce{CO2}` (:math:`c_a`) in part per million to Pascals,
    accounting for atmospheric pressure.

    Args:
        co2: atmospheric :math:`\ce{CO2}`, ppm
        patm (float): atmospheric pressure, Pa

    Returns:
        Ambient :math:`\ce{CO2}` in units of Pa

    Examples:
        >>> np.round(calc_co2_to_ca(413.03, 101325), 6)
        np.float64(41.850265)
    """

    return 1.0e-6 * co2 * patm  # Pa, atms. CO2



"""The :mod:`~pyrealm.constants` module implements the base
:class:`~pyrealm.constants.base.ConstantsClass` class and then implements the following
constants classes for different modules:

* :class:`~pyrealm.constants.competition_const.C3C4Const`
* :class:`~pyrealm.constants.core_const.CoreConst`
* :class:`~pyrealm.constants.isotope_const.IsotopesConst`
* :class:`~pyrealm.constants.pmodel_const.PModelConst`
* :class:`~pyrealm.constants.tmodel_const.TModelTraits`
* :class:`~pyrealm.constants.phenology_const.PhenologyConst`
"""  # noqa: D205, D415

from pyrealm.constants.base import ConstantsClass
from pyrealm.constants.competition_const import C3C4Const
from pyrealm.constants.core_const import CoreConst
from pyrealm.constants.isotope_const import IsotopesConst
from pyrealm.constants.phenology_const import PhenologyConst
from pyrealm.constants.pmodel_const import PModelConst
from pyrealm.constants.tmodel_const import TModelTraits

__all__ = [
    "C3C4Const",
    "ConstantsClass",
    "CoreConst",
    "IsotopesConst",
    "PModelConst",
    "PhenologyConst",
    "TModelTraits",
]



"""A set of functions implementing the crown shape and vertical leaf distribution model
used in PlantFATE :cite:t:`joshi:2022a`.
"""  # noqa: D205

from dataclasses import InitVar, dataclass, field
from typing import ClassVar

import numpy as np
from numpy.typing import NDArray

from pyrealm.core.experimental import warn_experimental
from pyrealm.demography.core import (
    PandasExporter,
    _validate_demography_array_arguments,
)
from pyrealm.demography.flora import Flora, StemTraits
from pyrealm.demography.tmodel import StemAllometry


def calculate_relative_crown_radius_at_z(
    z: NDArray[np.float64],
    stem_height: NDArray[np.float64],
    m: NDArray[np.float64],
    n: NDArray[np.float64],
    validate: bool = True,
    clip: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate relative crown radius at a given height.

    The crown shape parameters ``m`` and ``n`` define the vertical distribution of
    crown along the stem. For a stem of a given total height, this function calculates
    the relative crown radius at a given height :math:`z`:

    .. math::

        q(z) = m n \left(\dfrac{z}{H}\right) ^ {n -1}
        \left( 1 - \left(\dfrac{z}{H}\right) ^ n \right)^{m-1}

    This function calculates :math:`q(z)` across a set of stems: the ``stem_height``,
    ``m`` and ``n`` arguments should be one-dimensional arrays ('row vectors') of equal
    length :math:`I`.  The value for ``z`` is then an array of heights, with one of the
    following shapes:

    1. A scalar array: :math:`q(z)` is found for all stems at the same height and the
       return value is a 1D array of length :math:`I`.
    2. A row vector of length :math:`I`: :math:`q(z)` is found for all stems at
       stem-specific heights and the return value is again a 1D array of length
       :math:`I`.
    3. A column vector of length :math:`J`, that is a 2 dimensional array of shape
       (:math:`J`, 1). This allows :math:`q(z)` to be calculated efficiently for a set
       of heights for all stems and return a 2D array of shape (:math:`J`, :math:`I`).

    By default, this function clips :math:`q(z)`: the value is set to zero for values of
    :math:`z < 0` or :math:`z > H`.

    Args:
        z: Height at which to calculate relative radius
        stem_height: Total height of individual stem
        m: Canopy shape parameter of PFT
        n: Canopy shape parameter of PFT
        validate: Boolean flag to suppress argument validation.
        clip: Boolean flag to set :math:`q(z) = 0` where the :math:`z` is below zero or
            above the stem height.
    """

    if validate:
        _validate_demography_array_arguments(
            trait_args={"m": m, "n": n}, size_args={"stem_height": stem_height, "z": z}
        )

    z_over_height = z / stem_height
    q_z = m * n * z_over_height ** (n - 1) * (1 - z_over_height**n) ** (m - 1)

    # Set predictions to zero where z is below zero or above the stem height.
    if clip:
        q_z = np.where(np.logical_and(z >= 0, z <= stem_height), q_z, 0)

    return q_z


def calculate_crown_radius(
    q_z: NDArray[np.float64],
    r0: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate crown radius from relative crown radius and crown r0.

    The relative crown radius (:math:`q(z)`) at a given height :math:`z` describes the
    vertical profile of the crown shape, but only varies with the ``m`` and ``n`` shape
    parameters and the stem height. The actual crown radius at a given height
    (:math:`r(z)`) needs to be scaled using :math:`r_0` such that the maximum crown area
    equals the expected crown area given the crown area ratio traiit for the plant
    functional type:

    .. math::

        r(z) = r_0 q(z)

    This function calculates :math:`r(z)` given estimated ``r0`` and an array of
    relative radius values.

    Args:
        q_z: An array of relative crown radius values
        r0:  An array of crown radius scaling factor values
        validate: Boolean flag to suppress argument validation.
    """

    # TODO - think about validation here. qz must be row array or 2D (N, n_pft)

    return r0 * q_z


def calculate_stem_projected_crown_area_at_z(
    z: NDArray[np.float64],
    q_z: NDArray[np.float64],
    stem_height: NDArray[np.float64],
    crown_area: NDArray[np.float64],
    q_m: NDArray[np.float64],
    z_max: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    """Calculate stem projected crown area above a given height.

    This function calculates the projected crown area of a set of stems with given
    properties at a set of vertical heights. The stem properties are given in the
    arguments ``stem_height``,``crown_area``,``q_m`` and ``z_max``, which must be
    one-dimensional arrays ('row vectors') of equal length. The array of vertical
    heights ``z`` accepts a range of input shapes (see
    :meth:`~pyrealm.demography.crown.calculate_relative_crown_radius_at_z`
    ) and this function then also requires the expected relative stem radius (``q_z``)
    calculated from those heights.

    Args:
        z: Vertical height at which to estimate crown area
        q_z: Relative crown radius at those heights
        crown_area: Crown area of each stem
        stem_height: Stem height of each stem
        q_m: Canopy shape parameter ``q_m``` for each stem
        z_max: Height of maximum crown radius for each stem
        validate: Boolean flag to suppress argument validation.
    """

    if validate:
        _validate_demography_array_arguments(
            trait_args={"q_m": q_m},
            size_args={
                "stem_height": stem_height,
                "crown_area": crown_area,
                "z": z,
                "z_max": z_max,
            },
            at_size_args={"q_z": q_z},
        )

    # Calculate A_p
    # Calculate Ap given z > zm
    A_p = crown_area * (q_z / q_m) ** 2
    # Set Ap = Ac where z <= zm
    A_p = np.where(z <= z_max, crown_area, A_p)
    # Set Ap = 0 where z > H
    A_p = np.where(z > stem_height, 0, A_p)

    return A_p


def calculate_stem_projected_leaf_area_at_z(
    z: NDArray[np.float64],
    q_z: NDArray[np.float64],
    stem_height: NDArray[np.float64],
    crown_area: NDArray[np.float64],
    f_g: NDArray[np.float64],
    q_m: NDArray[np.float64],
    z_max: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    """Calculate projected leaf area above a given height.

    This function calculates the projected leaf area of a set of stems with given
    properties at a set of vertical heights. This differs from crown area in allowing
    for crown openness within the crown of an individual stem that results in the
    displacement of leaf area further down into the crown. The degree of openness is
    controlled by the crown gap fraction property of each stem.

    The stem properties are given in the arguments
    ``stem_height``,``crown_area``,``f_g``,``q_m`` and ``z_max``, which must be
    one-dimensional arrays ('row vectors') of equal length. The array of vertical
    heights ``z`` accepts a range of input shapes (see
    :meth:`~pyrealm.demography.crown.calculate_relative_crown_radius_at_z`
    ) and this function then also requires the expected relative stem radius (``q_z``)
    calculated from those heights.

    Args:
        z: Vertical heights on the z axis.
        q_z: Relative crown radius at heights in z.
        crown_area: Crown area for a stem
        stem_height: Total height of a stem
        f_g: Within crown gap fraction for each stem.
        q_m: Canopy shape parameter ``q_m``` for each stem
        z_max: Height of maximum crown radius for each stem
        validate: Boolean flag to suppress argument validation.
    """

    # NOTE: Although the internals of this function overlap a lot with
    #       calculate_stem_projected_crown_area_at_z, we want that function to be as
    #       lean as possible, as it used within solve_community_projected_crown_area.

    if validate:
        _validate_demography_array_arguments(
            trait_args={"q_m": q_m, "f_g": f_g},
            size_args={
                "stem_height": stem_height,
                "crown_area": crown_area,
                "z": z,
                "z_max": z_max,
            },
            at_size_args={"q_z": q_z},
        )

    # Calculate Ac terms
    A_c_terms = crown_area * (q_z / q_m) ** 2

    # Set Acp either side of z_max
    A_cp = np.where(
        z <= z_max,
        crown_area - A_c_terms * f_g,
        A_c_terms * (1 - f_g),
    )
    # Set Ap = 0 where z > H
    A_cp = np.where(z > stem_height, 0, A_cp)

    return A_cp


@dataclass
class CrownProfile(PandasExporter):
    """Calculate vertical crown profiles for stems.

    This method calculates crown profile predictions, given an array of vertical
    heights (``z``) for:

    * relative crown radius,
    * actual crown radius,
    * projected crown area, and
    * projected leaf area.

    The predictions require a set of plant functional types (PFTs) but also the expected
    allometric predictions of stem height, crown area and z_max for an actual stem of a
    given size for each PFT.

    In addition to the variables above, the class can also has properties the calculate
    the projected crown radius and projected leaf radius. These are simply the radii
    that would result in the two projected areas: the values are not directly meaningful
    for calculating canopy models, but can be useful for exploring the behavour of
    projected area on the same linear scale as the crown radius.

    Args:
        stem_traits: A Flora or StemTraits instance providing plant functional trait
            data.
        stem_allometry: A StemAllometry instance setting the stem allometries for the
            crown profile.
        z: An array of vertical height values at which to calculate crown profiles.
        validate: Boolean flag to suppress argument validation.
    """

    array_attrs: ClassVar[tuple[str, ...]] = (
        "relative_crown_radius",
        "crown_radius",
        "projected_crown_area",
        "projected_leaf_area",
        "projected_crown_radius",
        "projected_leaf_radius",
    )

    stem_traits: InitVar[StemTraits | Flora]
    """A Flora or StemTraits instance providing plant functional trait data."""
    stem_allometry: InitVar[StemAllometry]
    """A StemAllometry instance setting the stem allometries for the crown profile."""
    z: NDArray[np.float64]
    """An array of vertical height values at which to calculate crown profiles."""
    validate: InitVar[bool] = True
    """Boolean flag to suppress argument validation."""

    relative_crown_radius: NDArray[np.float64] = field(init=False)
    """An array of the relative crown radius of stems at z heights"""
    crown_radius: NDArray[np.float64] = field(init=False)
    """An array of the actual crown radius of stems at z heights"""
    projected_crown_area: NDArray[np.float64] = field(init=False)
    """An array of the projected crown area of stems at z heights"""
    projected_leaf_area: NDArray[np.float64] = field(init=False)
    """An array of the projected leaf area of stems at z heights"""

    # Information attributes
    _n_pred: int = field(init=False)
    """The number of predictions per stem."""
    _n_stems: int = field(init=False)
    """The number of stems."""

    __experimental__ = True

    def __post_init__(
        self,
        stem_traits: StemTraits | Flora,
        stem_allometry: StemAllometry,
        validate: bool,
    ) -> None:
        """Populate crown profile attributes from the traits, allometry and height."""

        warn_experimental("CrownProfile")

        # If validation is required, only need to perform validation once to check that
        # the at_dbh values are congruent with the stem_traits inputs. If they are, then
        # all the other allometry function inputs will be too.
        if validate:
            _validate_demography_array_arguments(
                trait_args={"h_max": stem_traits.h_max}, size_args={"z": self.z}
            )

        # Calculate relative crown radius
        self.relative_crown_radius = calculate_relative_crown_radius_at_z(
            z=self.z,
            m=stem_traits.m,
            n=stem_traits.n,
            stem_height=stem_allometry.stem_height,
            validate=False,
        )

        # Calculate actual radius
        self.crown_radius = calculate_crown_radius(
            q_z=self.relative_crown_radius, r0=stem_allometry.crown_r0, validate=False
        )

        # Calculate projected crown area
        self.projected_crown_area = calculate_stem_projected_crown_area_at_z(
            z=self.z,
            q_z=self.relative_crown_radius,
            crown_area=stem_allometry.crown_area,
            q_m=stem_traits.q_m,
            stem_height=stem_allometry.stem_height,
            z_max=stem_allometry.crown_z_max,
            validate=False,
        )

        # Calculate projected leaf area
        self.projected_leaf_area = calculate_stem_projected_leaf_area_at_z(
            z=self.z,
            q_z=self.relative_crown_radius,
            f_g=stem_traits.f_g,
            q_m=stem_traits.q_m,
            crown_area=stem_allometry.crown_area,
            stem_height=stem_allometry.stem_height,
            z_max=stem_allometry.crown_z_max,
            validate=False,
        )

        # Set the number of observations per stem (one if dbh is 1D, otherwise size of
        # the first axis)
        if self.relative_crown_radius.ndim == 1:
            self._n_pred = 1
        else:
            self._n_pred = self.relative_crown_radius.shape[0]

        self._n_stems = stem_traits._n_stems

    def __repr__(self) -> str:
        return (
            f"CrownProfile: Prediction for {self._n_stems} stems "
            f"at {self._n_pred} observations."
        )

    @property
    def projected_crown_radius(self) -> NDArray[np.float32]:
        """An array of the projected crown radius of stems at z heights."""
        return np.sqrt(self.projected_crown_area / np.pi)

    @property
    def projected_leaf_radius(self) -> NDArray[np.float32]:
        """An array of the projected leaf radius of stems at z heights."""
        return np.sqrt(self.projected_leaf_area / np.pi)


def get_crown_xy(
    crown_profile: CrownProfile,
    stem_allometry: StemAllometry,
    attr: str,
    stem_offsets: NDArray[np.float32] | None = None,
    two_sided: bool = True,
    as_xy: bool = False,
) -> list[tuple[NDArray, NDArray]] | list[NDArray]:
    """Extract plotting data from crown profiles.

    A CrownProfile instance contains crown radius and projected area data for a set of
    stems at given heights, but can contain predictions of these attributes above the
    actual heights of some or all of the stems or indeed below ground.

    This function extracts plotting data for a given attribute for each crown that
    includes only the predictions within the height range of the actual stem. It can
    also mirror the values around the vertical midline to provide a two sided canopy
    shape.

    The data are returned as a list with one entry per stem. The default value for each
    entry a tuple of two arrays (height, attribute values) but the `as_xy=True` option
    will return an `(N, 2)` dimensioned XY array suitable for use with
    {class}`~matplotlib.patches.Polygon`.

    Args:
        crown_profile: A crown profile instance
        stem_allometry: The stem allometry instance used to create the crown profile
        attr: The crown profile attribute to plot (see
            :class:`~pyrealm.demography.crown.CrownProfile`)
        stem_offsets: An optional array of offsets to add to the midline of stems.
        two_sided: Should the plotting data show a two sided canopy.
        as_xy: Should the plotting data be returned as a single XY array.

    """

    # Input validation
    if attr not in crown_profile.array_attrs:
        raise ValueError(f"Unknown crown profile attribute: {attr}")

    # Get the attribute and flatten the heights from a column array to one dimensional
    attr_values = getattr(crown_profile, attr)
    z = crown_profile.z.flatten()

    # Orient the data so that lower heights always come first
    if z[0] < z[-1]:
        z = np.flip(z)
        attr_values = np.flip(attr_values, axis=0)

    # Collect the per stem data
    crown_plotting_data: list = []

    for stem_index in np.arange(attr_values.shape[1]):
        # Find the heights and values that fall within the individual stem
        height_is_valid = np.logical_and(
            z <= stem_allometry.stem_height[:, stem_index], z >= 0
        )
        valid_attr_values: NDArray = attr_values[height_is_valid, stem_index]
        valid_heights: NDArray = z[height_is_valid]

        if two_sided:
            # The values are extended to include the reverse profile as well as the zero
            # value at the stem height
            valid_heights = np.concatenate(
                [
                    np.flip(valid_heights),
                    stem_allometry.stem_height[:, stem_index],
                    valid_heights,
                ]
            )
            valid_attr_values = np.concatenate(
                [-np.flip(valid_attr_values), [0], valid_attr_values]
            )
        else:
            # Only the zero value is added
            valid_heights = np.concatenate(
                [
                    stem_allometry.stem_height[:, stem_index],
                    valid_heights,
                ]
            )
            valid_attr_values = np.concatenate([[0], valid_attr_values])

        # Add offsets if provided
        if stem_offsets is not None:
            valid_attr_values += stem_offsets[stem_index]

        if as_xy:
            # Combine the values into an (N,2) XY array
            crown_plotting_data.append(
                np.hstack([valid_attr_values[:, None], valid_heights[:, None]])
            )
        else:
            # Return the individual 1D arrays
            crown_plotting_data.append((valid_heights, valid_attr_values))

    return crown_plotting_data



"""This modules provides the Community class, which contains the set of size-structured
cohorts of plants across a range of plant functional types that occur a given location
(or 'cell') with a given cell id number and area.

The class provides factory methods to create Community instances from CSV, JSON and TOML
files, using :mod:`marshmallow` schemas to both validate the input data and to perform
post processing to align the input formats to the initialisation arguments to the
Community class.

Internally, the cohort data in the Community class is represented as a dictionary of
`numpy` arrays.

Worked example
==============

The example code below demonstrates defining PFTs, creating a Flora collection,
initializing a Community, and computing ecological metrics using the T Model for a set
of plant cohorts.

>>> import pandas as pd
>>>
>>> from pyrealm.demography.flora import PlantFunctionalType, Flora
>>> from pyrealm.demography.tmodel import (
...     calculate_heights, calculate_crown_areas, calculate_stem_masses,
...     calculate_foliage_masses
... )

>>> pft1 = PlantFunctionalType(
...     name="Evergreen Tree",
...     a_hd=120.0,
...     ca_ratio=380.0,
...     h_max=30.0,
...     rho_s=210.0,
...     lai=3.0,
...     sla=12.0,
...     tau_f=5.0,
...     tau_r=1.2,
...     par_ext=0.6,
...     yld=0.65,
...     zeta=0.18,
...     resp_r=0.95,
...     resp_s=0.045,
...     resp_f=0.12,
...     m=2.5,
...     n=4.5,
... )

>>> pft2 = PlantFunctionalType(
...     name="Deciduous Shrub",
...     a_hd=100.0,
...     ca_ratio=350.0,
...     h_max=4.0,
...     rho_s=180.0,
...     lai=2.0,
...     sla=15.0,
...     tau_f=3.0,
...     tau_r=0.8,
...     par_ext=0.4,
...     yld=0.55,
...     zeta=0.15,
...     resp_r=0.85,
...     resp_s=0.05,
...     resp_f=0.1,
...     m=3.0,
...     n=5.0,
... )

Create a Flora collection:

>>> flora = Flora([pft1, pft2])

Define community data as size-structured cohorts of given plant functional types with a
given number of individuals.

>>> cohort_dbh_values = np.array([0.10, 0.03, 0.12, 0.025])
>>> cohort_n_individuals = np.array([100, 200, 150, 180])
>>> cohort_pft_names = np.array(
...    ["Evergreen Tree", "Deciduous Shrub", "Evergreen Tree", "Deciduous Shrub"]
... )

Initialize a Community into an area of 1000 square meter with the given cohort data:

>>> community = Community(
...     cell_id=1,
...     cell_area=1000.0,
...     flora=flora,
...     cohorts=Cohorts(
...         dbh_values=cohort_dbh_values,
...         n_individuals=cohort_n_individuals,
...         pft_names=cohort_pft_names,
...     ),
... )

The data in the Community class is stored under three attributes, each of which stores
an instance of a dataclass holding related parts of the community data. All have a
``to_pandas`` method that can be used to visualise and explore the data:

>>> community.cohorts.to_pandas()
   dbh_values  n_individuals        pft_names
0       0.100            100   Evergreen Tree
1       0.030            200  Deciduous Shrub
2       0.120            150   Evergreen Tree
3       0.025            180  Deciduous Shrub

>>> community.stem_allometry.to_pandas()[
...     ["stem_height", "crown_area", "stem_mass", "crown_r0", "crown_z_max"]
... ]  # doctest: +NORMALIZE_WHITESPACE
                   stem_height  crown_area  stem_mass  crown_r0  crown_z_max
column_stem_index
0                     9.890399    2.459835   8.156296  0.339477     7.789552
1                     2.110534    0.174049   0.134266  0.083788     1.642777
2                    11.436498    3.413238  13.581094  0.399890     9.007241
3                     1.858954    0.127752   0.082126  0.071784     1.446955

>>> community.stem_traits.to_pandas()[
...     ["name", "a_hd", "ca_ratio", "sla", "par_ext", "q_m",  "z_max_prop"]
... ]
              name   a_hd  ca_ratio   sla  par_ext       q_m  z_max_prop
0   Evergreen Tree  120.0     380.0  12.0      0.6  2.606561    0.787587
1  Deciduous Shrub  100.0     350.0  15.0      0.4  2.809188    0.778371
2   Evergreen Tree  120.0     380.0  12.0      0.6  2.606561    0.787587
3  Deciduous Shrub  100.0     350.0  15.0      0.4  2.809188    0.778371
"""  # noqa: D205

from __future__ import annotations

import json
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, ClassVar

import numpy as np
import pandas as pd
from marshmallow import Schema, fields, post_load, validate, validates_schema
from marshmallow.exceptions import ValidationError
from numpy.typing import NDArray

from pyrealm.core.experimental import warn_experimental
from pyrealm.core.utilities import check_input_shapes
from pyrealm.demography.core import CohortMethods, PandasExporter
from pyrealm.demography.flora import Flora, StemTraits
from pyrealm.demography.tmodel import StemAllometry

if sys.version_info[:2] >= (3, 11):
    import tomllib
    from tomllib import TOMLDecodeError
else:
    import tomli as tomllib
    from tomli import TOMLDecodeError


@dataclass
class Cohorts(PandasExporter, CohortMethods):
    """A dataclass to hold data for a set of plant cohorts.

    The attributes should be numpy arrays of equal length, containing an entry for each
    cohort in the data class.
    """

    # A class variable setting the attribute names of traits.
    array_attrs: ClassVar[tuple[str, ...]] = tuple(
        ["dbh_values", "n_individuals", "pft_names"]
    )
    count_attr: ClassVar[str] = "n_cohorts"

    # Instance attributes
    dbh_values: NDArray[np.float64]
    n_individuals: NDArray[np.int_]
    pft_names: NDArray[np.str_]
    n_cohorts: int = field(init=False)

    __experimental__ = True

    def __post_init__(self) -> None:
        """Validation of cohorts data."""

        # TODO - validation - maybe make this optional to reduce workload within
        # simulations

        warn_experimental("Cohorts")

        # Check cohort data types
        if not (
            isinstance(self.dbh_values, np.ndarray)
            and isinstance(self.n_individuals, np.ndarray)
            and isinstance(self.pft_names, np.ndarray)
        ):
            raise ValueError("Cohort data not passed as numpy arrays")

        # Check the cohort inputs are of equal length
        try:
            check_input_shapes(self.dbh_values, self.n_individuals, self.dbh_values)
        except ValueError:
            raise ValueError("Cohort arrays are of unequal length")

        self.n_cohorts = len(self.dbh_values)


class CohortSchema(Schema):
    """A validation schema for Cohort data objects.

    This schema can be used to validate the ``cohorts`` components of JSON and TOML
    community data files, which are simple dictionaries:

    .. code-block:: toml
        :caption: TOML

        dbh_value = 0.2
        n_individuals = 6
        pft_name = "broadleaf"

    .. code-block:: json
        :caption: JSON

        {
            "pft_name": "broadleaf",
            "dbh_value": 0.2,
            "n_individuals": 6
        }
    """

    dbh_value = fields.Float(
        required=True, validate=validate.Range(min=0, min_inclusive=False)
    )
    n_individuals = fields.Integer(
        strict=True, required=True, validate=validate.Range(min=0, min_inclusive=False)
    )
    pft_name = fields.Str(required=True)


class CommunityStructuredDataSchema(Schema):
    """A validation schema for Cohort data in a structured format (JSON/TOML).

    This schema can be used to validate data for creating a Community instance stored in
    a structured format such as JSON or TOML. The format is expected to provide a cell
    area and id along with an array of cohort objects providing the plant functional
    type name, diameter at breast height (DBH) and number of individuals (see
    :class:`~pyrealm.demography.community.CohortSchema`). Example inputs with this
    structure are:

    .. code-block:: toml
        :caption: TOML

        cell_area = 100
        cell_id = 1

        [[cohorts]]
        dbh_value = 0.2
        n_individuals = 6
        pft_name = "broadleaf"

        [[cohorts]]
        dbh_value = 0.25
        n_individuals = 6
        pft_name = "conifer"

    .. code-block:: json
        :caption: JSON

        {
        "cell_id": 1,
        "cell_area": 100,
        "cohorts": [
            {
                "pft_name": "broadleaf",
                "dbh_value": 0.2,
                "n_individuals": 6
            },
            {
                "pft_name": "broadleaf",
                "dbh_value": 0.25,
                "n_individuals": 6
            }]
        }

    Any data validated with this schema is post-processed to convert the cohort objects
    into the arrays of cohort data required to initialise instances of the
    :class:`~pyrealm.demography.community.Community` class.
    """

    cell_id = fields.Integer(required=True, strict=True, validate=validate.Range(min=0))
    cell_area = fields.Float(
        required=True, validate=validate.Range(min=0, min_inclusive=False)
    )
    cohorts = fields.List(
        fields.Nested(CohortSchema), required=True, validate=validate.Length(min=1)
    )

    @post_load
    def convert_to_community_args(self, data: dict, **kwargs: Any) -> dict[str, Any]:
        """Convert cohorts to arrays.

        This post load method converts the cohort arrays into a Cohorts objects and
        packages the data up into the required arguments used to initialise a Community
        object.

        Args:
            data: Data passed to the validator
            kwargs: Additional keyword arguments passed by marshmallow
        """

        return dict(
            cell_id=data["cell_id"],
            cell_area=data["cell_area"],
            cohorts=Cohorts(
                dbh_values=np.array([c["dbh_value"] for c in data["cohorts"]]),
                n_individuals=np.array([c["n_individuals"] for c in data["cohorts"]]),
                pft_names=np.array([c["pft_name"] for c in data["cohorts"]]),
            ),
        )


class CommunityCSVDataSchema(Schema):
    """A validation schema for community initialisation data in CSV format.

    This schema can be used to validate data for creating a Community instance stored in
    CSV format. The file is expected to provide fields providing cell id and cell area
    and then functional type name, diameter at breast height (DBH) and number of
    individuals. Each row is taken to represent a cohort and the cell id and area
    *must** be consistent across rows.

    .. code-block::

        cell_id,cell_area,cohort_pft_names,cohort_dbh_values,cohort_n_individuals
        1,100,broadleaf,0.2,6
        1,100,broadleaf,0.25,6
        1,100,broadleaf,0.3,3
        1,100,broadleaf,0.35,1
        1,100,conifer,0.5,1
        1,100,conifer,0.6,1

    The input data is expected to be provided to this schema as a dictionary of lists of
    field values keyed by field name, as for example by using
    :meth:`pandas.DataFrame.to_dict` with the ``orient='list'`` argument.

    The schema automatically validates that the cell id and area are consistent and then
    post-processing is used to simplify those fields to the scalar inputs required to
    initialise instances of the  :class:`~pyrealm.demography.community.Community` class
    and to convert the cohort data into arrays,
    """

    cell_id = fields.List(fields.Integer(strict=True), required=True)
    cell_area = fields.List(fields.Float(), required=True)
    cohort_dbh_values = fields.List(fields.Float(), required=True)
    cohort_n_individuals = fields.List(fields.Integer(strict=True), required=True)
    cohort_pft_names = fields.List(fields.Str(), required=True)

    @validates_schema
    def validate_consistent_cell_data(self, data: dict, **kwargs: Any) -> None:
        """Schema wide validation.

        Args:
            data: Data passed to the validator
            kwargs: Additional keyword arguments passed by marshmallow
        """

        # Check cell area and cell id consistent
        if not all([c == data["cell_id"][0] for c in data["cell_id"]]):
            raise ValueError(
                "Multiple cell id values fields in community data, see load_communities"
            )

        if not all([c == data["cell_area"][0] for c in data["cell_area"]]):
            raise ValueError("Cell area varies in community data")

    @post_load
    def convert_to_community_args(self, data: dict, **kwargs: Any) -> dict[str, Any]:
        """Make cell data scalar.

        This post load method reduces the repeated cell id and cell area across CSV data
        rows into the scalar inputs required to initialise a Community object and
        packages the data on individual cohorts into a Cohorts object.
        """

        return dict(
            cell_id=data["cell_id"][0],
            cell_area=data["cell_area"][0],
            cohorts=Cohorts(
                dbh_values=np.array(data["cohort_dbh_values"]),
                n_individuals=np.array(data["cohort_n_individuals"]),
                pft_names=np.array(data["cohort_pft_names"]),
            ),
        )


@dataclass
class Community:
    """The plant community class.

    A community is a set of size-structured plant cohorts in a given location, where the
    location has a specified numeric id and a known area in square meters.

    A cohort defines a number of individual plants with the same diameter at breast
    height (DBH) and plant functional type (PFT). Internally, the cohort data is built
    into a :class:`pandas.DataFrame` with each row representing a cohort and each column
    representing a property of the cohort. The initial input data is extended to include
    the plant functional type traits for each cohort (see
    :class:`~pyrealm.demography.flora.PlantFunctionalType`) and then is further extended
    to include the geometric and canopy predictions of the T Model for each cohort.

    Factory methods are provided to load community data from csv, TOML or JSON files.

    Args:
        cell_id: An positive integer id for the community location.
        cell_area: An area in square metres for the community location.
        flora: A flora object containing the plant functional types for the community
        cohort_dbh_values: A numpy array giving the diameter at breast height in metres
            for each cohort.
        cohort_n_individuals: A numpy array giving the number of individuals in each
            cohort.
        cohort_pft_names: A numpy array giving the name of the plant functional type in
            each cohort.
    """

    # Dataclass attributes for initialisation
    # - community wide properties
    cell_id: int
    cell_area: float
    flora: Flora

    # - arrays representing properties of cohorts
    cohorts: Cohorts

    # Post init properties
    n_cohorts: int = field(init=False)
    stem_traits: StemTraits = field(init=False)
    stem_allometry: StemAllometry = field(init=False)

    __experimental__ = True

    def __post_init__(
        self,
    ) -> None:
        """Validate inputs and populate derived community attributes.

        The ``__post_init__`` builds a pandas dataframe of PFT values and T model
        predictions across the validated initial cohort data.
        """

        warn_experimental("Community")

        # Check cell area and cell id
        if not (isinstance(self.cell_area, float | int) and self.cell_area > 0):
            raise ValueError("Community cell area must be a positive number.")

        if not (isinstance(self.cell_id, int) and self.cell_id >= 0):
            raise ValueError("Community cell id must be a integer >= 0.")

        # How many cohorts
        self.n_cohorts = self.cohorts.n_cohorts

        # Get the stem traits for the cohorts
        self.stem_traits = self.flora.get_stem_traits(self.cohorts.pft_names)

        # Populate the stem allometry
        self.stem_allometry = StemAllometry(
            stem_traits=self.stem_traits, at_dbh=self.cohorts.dbh_values
        )

    @classmethod
    def from_csv(cls, path: Path, flora: Flora) -> Community:
        """Create a Community object from a CSV file.

        This factory method checks that the required fields are present in the CSV data
        and that the cell_id and cell_area values are constant. It then passes the data
        through further validation using the
        `meth`:`~pyrealm.demography.community.Community._from_file_data` method and
        returns a Community instance.

        Args:
            path: A path to a CSV file of community data
            flora: A Flora instance providing plant functional types used in the
                community data
        """

        # Load the data
        try:
            file_data = pd.read_csv(path)
        except (FileNotFoundError, pd.errors.ParserError) as excep:
            raise excep

        # Validate the data - there is an inconsequential typing issue here:
        # Argument "data" to "load" of "Schema" has incompatible type
        # "dict[Hashable, Any]";
        # expected "Mapping[str, Any] | Iterable[Mapping[str, Any]]"
        try:
            file_data = CommunityCSVDataSchema().load(data=file_data.to_dict("list"))  # type: ignore[arg-type]
        except ValidationError as excep:
            raise excep

        return cls(**file_data, flora=flora)

    @classmethod
    def from_json(cls, path: Path, flora: Flora) -> Community:
        """Create a Community object from a JSON file.

        This factory method loads community data from a JSON community file and
        validates it using
        `class`:`~pyrealm.demography.community.CommunityStructuredDataSchema` before
        using the data to initialise a Community instance.

        Args:
            path: A path to a JSON file of community data
            flora: A Flora instance providing plant functional types used in the
                community data
        """

        # Load the data
        try:
            file_data = json.load(open(path))
        except (FileNotFoundError, json.JSONDecodeError) as excep:
            raise excep

        # Validate the data
        try:
            file_data = CommunityStructuredDataSchema().load(data=file_data)
        except ValidationError as excep:
            raise excep

        return cls(**file_data, flora=flora)

    @classmethod
    def from_toml(cls, path: Path, flora: Flora) -> Community:
        """Create a Community object from a TOML file.

        This factory method loads community data from a TOML community file and
        validates it using
        `class`:`~pyrealm.demography.community.CommunityStructuredDataSchema` before
        using the data to initialise a Community instance.

        Args:
            path: A path to a TOML file of community data
            flora: A Flora instance providing plant functional types used in the
                community data
        """

        # Load the data
        try:
            file_data = tomllib.load(open(path, "rb"))
        except (FileNotFoundError, TOMLDecodeError) as excep:
            raise excep

        # Validate the data
        try:
            file_data = CommunityStructuredDataSchema().load(data=file_data)
        except ValidationError as excep:
            raise excep

        return cls(**file_data, flora=flora)

    def drop_cohorts(self, drop_indices: NDArray[np.int_]) -> None:
        """Drop cohorts from the community.

        This method drops the identified cohorts from the ``cohorts`` attribute and then
        removes their data from  the ``stem_traits`` and ``stem_allometry`` attributes
        to match.
        """

        self.cohorts.drop_cohort_data(drop_indices=drop_indices)
        self.stem_traits.drop_cohort_data(drop_indices=drop_indices)
        self.stem_allometry.drop_cohort_data(drop_indices=drop_indices)

        self.n_cohorts = self.cohorts.n_cohorts

    def add_cohorts(self, new_data: Cohorts) -> None:
        """Add a new set of cohorts to the community.

        This method extends the ``cohorts`` attribute with the new cohort data and then
        also extends the ``stem_traits`` and ``stem_allometry`` to match.

        Args:
            new_data: An instance of :class:`~pyrealm.demography.community.Cohorts`
                containing cohort data to add to the community.
        """

        self.cohorts.add_cohort_data(new_data=new_data)

        new_stem_traits = self.flora.get_stem_traits(pft_names=new_data.pft_names)
        self.stem_traits.add_cohort_data(new_data=new_stem_traits)

        new_stem_allometry = StemAllometry(
            stem_traits=new_stem_traits, at_dbh=new_data.dbh_values
        )
        self.stem_allometry.add_cohort_data(new_data=new_stem_allometry)

        self.n_cohorts = self.cohorts.n_cohorts

    # @classmethod
    # def load_communities_from_csv(
    #     cls, cell_area: float, csv_path: str, flora: Flora
    # ) -> list[Community]:
    #     """Loads a list of communities from a csv provided in the appropriate format.

    #     The csv should contain the following columns: cell_id,
    #     diameter_at_breast_height, plant_functional_type, number_of_individuals. Each
    #     row in the csv should represent one cohort.

    #     :param cell_area: the area of the cell at each location, this is assumed to be
    #     the same across all the locations in the csv.
    #     :param csv_path: path to the csv containing community data, as detailed above.
    #     :param flora: a flora object, ie a dictionary of plant functional properties,
    #     keyed by pft name.
    #     :return: a list of community objects, loaded from the csv
    #     file.
    #     """
    #     community_data = pd.read_csv(csv_path)

    #     data_grouped_by_community = community_data.groupby(community_data.cell_id)

    #     communities = []

    #     for cell_id in data_grouped_by_community.groups:
    #         community_dataframe = data_grouped_by_community.get_group(cell_id)
    #         dbh_values = community_dataframe["diameter_at_breast_height"].to_numpy(
    #             dtype=np.float32
    #         )
    #         number_of_individuals = community_dataframe[
    #             "number_of_individuals"
    #         ].to_numpy(dtype=np.int_)
    #         pft_names = community_dataframe["plant_functional_type"].to_numpy(
    #               dtype=str
    #         )
    #         community_object = Community(
    #             cell_id,  # type:ignore
    #             cell_area,
    #             dbh_values,
    #             number_of_individuals,
    #             pft_names,
    #             flora,
    #         )
    #         communities.append(community_object)

    #     return communities



"""The ``solar`` submodule provides the DailySolarFluxes class, used to calculate daily
solar radiation fluxes for observations.
"""  # noqa: D205

from dataclasses import InitVar, dataclass, field

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import CoreConst
from pyrealm.core.calendar import Calendar
from pyrealm.core.solar import (
    _calculate_daily_solar_radiation,
    _calculate_daytime_net_radiation,
    _calculate_net_radiation_crossover_hour_angle,
    _calculate_nighttime_net_radiation,
    _calculate_sunset_hour_angle,
    calculate_distance_factor,
    calculate_heliocentric_longitudes,
    calculate_net_longwave_radiation,
    calculate_ppfd_from_tau_rd,
    calculate_ru_rv_intermediates,
    calculate_rw_intermediate,
    calculate_solar_declination_angle,
    calculate_transmissivity,
)
from pyrealm.core.utilities import check_input_shapes


@dataclass
class DailySolarFluxes:
    """Calculate daily solar fluxes.

    This dataclass takes arrays describing the latitude, elevation, sunshine fraction
    and mean daily temperature for observations and then calculates key radiation fluxes
    given a Calendar object providing the Julian day of the observations and the year
    and number of days in the year.

    Args:
        latitude: The Latitude of observations (degrees)
        elevation: Elevation of observations (metres)
        dates: Dates of observations
        sunshine_fraction: Daily sunshine fraction of observations (unitless)
        temperature: Daily temperature of observations (°C)
    """

    latitude: InitVar[NDArray[np.float64]]
    elevation: InitVar[NDArray[np.float64]]
    dates: Calendar
    sunshine_fraction: InitVar[NDArray[np.float64]]
    temperature: InitVar[NDArray[np.float64]]
    core_const: CoreConst = field(default_factory=lambda: CoreConst())

    nu: NDArray[np.float64] = field(init=False)
    r"""True heliocentric anomaly (:math:`\nu`, degrees)"""
    lambda_: NDArray[np.float64] = field(init=False)
    r"""True heliocentric longitude, (:math:`\lambda`, degrees)"""
    distance_factor: NDArray[np.float64] = field(init=False)
    """Distance factor (:math:`d_r`, -)"""
    declination: NDArray[np.float64] = field(init=False)
    r"""Declination angle (:math:`\delta`, degrees)"""
    ru: NDArray[np.float64] = field(init=False)
    """Intermediate variable (:math:`r_u`, unitless)"""
    rv: NDArray[np.float64] = field(init=False)
    """Intermediate variable (:math:`r_v`, unitless)"""
    sunset_hour_angle: NDArray[np.float64] = field(init=False)
    """Sunset hour angle (:math:`h_s`, degrees)"""
    daily_solar_radiation: NDArray[np.float64] = field(init=False)
    """Daily extraterrestrial solar radiation (:math:`R_d`, J m-2)"""
    transmissivity: NDArray[np.float64] = field(init=False)
    r"""Transmittivity (:math:`\tau`, unitless)"""
    daily_ppfd: NDArray[np.float64] = field(init=False)
    """Daily photosynthetic photon flux density (PPFD, µmol m-2 s-1)"""
    net_longwave_radiation: NDArray[np.float64] = field(init=False)
    """Net longwave radiation (:math:`R_{nl}`, W m-2)"""
    rw: NDArray[np.float64] = field(init=False)
    """Intermediate variable (:math:`r_w`,  W m-2)"""
    crossover_hour_angle: NDArray[np.float64] = field(init=False)
    """Net radiation cross-over hour angle, (:math:`h_n`, degrees)"""
    daytime_net_radiation: NDArray[np.float64] = field(init=False)
    """Daytime net radiation (:math:`R_{d}`, J m-2)"""
    nighttime_net_radiation: NDArray[np.float64] = field(init=False)
    """Nighttime net radiation (:math:`R_{nn}`, J m-2)"""

    def __post_init__(
        self,
        latitude: NDArray[np.float64],
        elevation: NDArray[np.float64],
        sunshine_fraction: NDArray[np.float64],
        temperature: NDArray[np.float64],
    ) -> None:
        """Populates key fluxes from input variables."""

        # Validate the inputs
        shapes = check_input_shapes(latitude, elevation, sunshine_fraction, temperature)
        if len(self.dates) != shapes[0]:
            raise ValueError(
                "The calendar is not the same length as the first axis of inputs "
            )

        # Calculate heliocentric longitudes (nu and lambda), Berger (1978)
        nu, lambda_ = calculate_heliocentric_longitudes(
            ordinal_date=self.dates.julian_day, n_days=self.dates.days_in_year
        )

        # Calculate distance factor (dr), Berger et al. (1993)
        distance_factor = calculate_distance_factor(
            nu=nu, solar_eccentricity=self.core_const.solar_eccentricity
        )

        # Calculate declination angle (delta), Woolf (1968)
        delta = calculate_solar_declination_angle(
            lambda_=lambda_, solar_obliquity=self.core_const.solar_obliquity
        )

        # The nu, lambda_, distance_factor and declination attributes are all one
        # dimensional arrays calculated from the Calendar along the first (time) axis of
        # the other inputs. These need to be broadcastable to the shape of the other
        # inputs. The expand_dims variable gets a list of the axes to expand onto -
        # which will be an empty list when ndim=1, leaving the targets unchanged.
        expand_dims = list(np.arange(1, elevation.ndim))
        self.nu = np.expand_dims(nu, axis=expand_dims)
        self.lambda_ = np.expand_dims(lambda_, axis=expand_dims)
        self.distance_factor = np.expand_dims(distance_factor, axis=expand_dims)
        self.declination = np.expand_dims(delta, axis=expand_dims)

        # Calculate transmittivity (tau), unitless
        # Eq. 11, Linacre (1968); Eq. 2, Allen (1996)
        self.transmissivity = calculate_transmissivity(
            sunshine_fraction=sunshine_fraction,
            elevation=elevation,
            coef=self.core_const.transmissivity_coef,
        )

        # Calculate intermediate values ru, rv, rw
        self.ru, self.rv = calculate_ru_rv_intermediates(
            declination=self.declination, latitude=latitude
        )

        self.rw = calculate_rw_intermediate(
            transmissivity=self.transmissivity,
            distance_factor=self.distance_factor,
            shortwave_albedo=self.core_const.shortwave_albedo,
            solar_constant=self.core_const.solar_constant,
        )

        # Calculate the sunset hour angle (hs), Eq. 3.22, Stine & Geyer (2001)
        self.sunset_hour_angle = _calculate_sunset_hour_angle(ru=self.ru, rv=self.rv)

        # Calculate daily extraterrestrial solar radiation (R_d, J/m^2)
        # Eq. 1.10.3, Duffy & Beckman (1993)
        self.daily_solar_radiation = _calculate_daily_solar_radiation(
            ru=self.ru,
            rv=self.rv,
            distance_factor=self.distance_factor,
            sunset_hour_angle=self.sunset_hour_angle,
            day_seconds=self.core_const.day_seconds,
            solar_constant=self.core_const.solar_constant,
        )

        # Calculate daily PPFD (ppfd_d), mol/m^2
        self.daily_ppfd = calculate_ppfd_from_tau_rd(
            transmissivity=self.transmissivity,
            daily_solar_radiation=self.daily_solar_radiation,
            swdown_to_ppfd_factor=self.core_const.swdown_to_ppfd_factor,
            visible_light_albedo=self.core_const.visible_light_albedo,
        )

        # Estimate net longwave radiation (rnl), W/m^2
        # Eq. 11, Prentice et al. (1993); Eq. 5 and 6, Linacre (1968)
        self.net_longwave_radiation = calculate_net_longwave_radiation(
            sunshine_fraction=sunshine_fraction,
            temperature=temperature,
            coef=self.core_const.net_longwave_radiation_coef,
        )

        # Calculate net radiation cross-over hour angle (hn), degrees
        self.crossover_hour_angle = _calculate_net_radiation_crossover_hour_angle(
            ru=self.ru,
            rv=self.rv,
            rw=self.rw,
            net_longwave_radiation=self.net_longwave_radiation,
        )

        # Calculate daytime net radiation (rn_d), J/m^2
        self.daytime_net_radiation = _calculate_daytime_net_radiation(
            ru=self.ru,
            rv=self.rv,
            rw=self.rw,
            crossover_hour_angle=self.crossover_hour_angle,
            net_longwave_radiation=self.net_longwave_radiation,
            day_seconds=self.core_const.day_seconds,
        )

        # Calculate nighttime net radiation (rnn_d), J/m^2
        self.nighttime_net_radiation = _calculate_nighttime_net_radiation(
            ru=self.ru,
            rv=self.rv,
            rw=self.rw,
            net_longwave_radiation=self.net_longwave_radiation,
            crossover_hour_angle=self.crossover_hour_angle,
            sunset_hour_angle=self.sunset_hour_angle,
            day_seconds=self.core_const.day_seconds,
        )
